import augur
from augur.integrations.objects import JiraObject


class JiraPerson(JiraObject):

    def __init__(self, source, **kwargs):
        super(JiraPerson, self).__init__(source, **kwargs)
        self._jira_person = None
        self._db_person = None

    def _load(self):

        if not self.option('jira_username'):
            self.logger("Cannot initialize JiraPerson without Jira username")

        username = self.option('jira_username')
        self._db_person = augur.api.get_staff_member_by_field(username=username) or None

        if self._db_person:
            if self.option('load_jira_user_data'):
                self.log_access('load-user')

                user = self.source.jira.user(username)
                self._jira_person = user.raw()



from copy import copy

import arrow
import datetime
from arrow import Arrow
from jira import Issue
from munch import munchify

from augur.context import AugurContext
from augur.common import POSSIBLE_DATE_TIME_FORMATS
from augur.integrations.objects.base import JiraObject, InvalidId
from functools import reduce


class JiraIssue(JiraObject):
    """
    Represents a collection of issues

    Options:
        - key (Optional) - The key of the issue to load
    """

    def __init__(self, source, **kwargs):
        super(JiraIssue, self).__init__(source, **kwargs)
        self._issue = None
        self._epic = None
        self._parent = None
        self.default_fields = self.source.default_fields

    def _load(self):

        if not self.option("key"):
            raise InvalidId("Issue key must be given")

        self.log_access('issue',self.option('key'))
        issue = self.source.jira.issue(self.option('key'))
        if issue:
            self._issue = munchify(issue.raw)
            return True
        else:
            self._issue = None
            self.logger.error("JiraIssue: Unable to load issue from Jira")
            return False

    def __eq__(self, issue):
        """
        Override equality to use the issue key (when available) as the compared value as opposed to the object ID
        :param cmp: The JiraIssue object to cmpare against.
        :return:
        """
        if not self.key and not issue.key:
            # when both are empty we compare the object IDs instead
            return id(self) == id(issue)
        else:
            return self.key == issue.key

    def __ne__(self, issue):
        return not self.__eq__(issue)

    def __hash__(self):
        # use the issue key as the hash if the issue exists, otherwise fallback to using the python object id.
        return hash(self.key) if self.key else hash(id(self))

    @property
    def is_subtask(self):
        return self.get_field('issuetype.subtask') is not None

    @property
    def summary(self):
        return self.get_field('summary') or ""

    @property
    def status(self):
        return self.get_field('status.name') or ""

    @property
    def priority(self):
        return self.get_field('priority.name') or ""

    @property
    def issuetype(self):
        return self.get_field('issuetype.name') or ""

    @property
    def description(self):
        return self.get_field('description') or ""

    @property
    def reporter(self):
        return self.get_field('reporter.name') or ""

    @property
    def assignee(self):
        return self.get_field('assignee.name') or ""

    @property
    def points(self):
        return self.get_field("story points", translate=True) or 0.0

    @property
    def resolution(self):
        return self.get_field('resolution.name') or ""

    @property
    def key(self):
        return self.get_field('key') or ""

    @property
    def team_name(self):
        return self.get_field('dev team.value', translate=True) or ""

    @property
    def issue(self):
        """
        Returns the issue as a Munch dictionary object.  Meaning that you can access its properties using dot
        notation or dict notation.  This is the raw issue as returned from Jira otherwise.
        :return: The issue object as a Munch instance
        """
        return self._issue

    def get_parent(self, only_key=True):
        """
        Gets the parent key or issue object
        :param only_key: If True then only returns the key as a string.  Otherwise returns a new JiraIssue object
                            containing parent information
        :return: Return either a parent issue key, the parent issue JiraIssue object or None if not found.
        """
        parent_key = self.get_field('parent.key')
        if not only_key and parent_key:
            if self._parent:
                return self._parent

            self._parent = JiraIssue(source=self.source,key=parent_key)
            if self._parent.load():
                return self._parent
            else:
                return None

        return parent_key

    def get_epic(self, only_key=True):
        """
        Gets the associated epic key or issue object
        :param only_key: If True then only returns the key as a string.  Otherwise returns a new JiraIssue object
                            containing epic information
        :return: Return either an epic issue key, the epic JiraIssue object or None if not found.
        """

        field_name = self.default_fields['epic link']
        epic_key = self.get_field('fields.%s'%field_name)

        if epic_key and not only_key:
            if self._epic:
                return self._epic

            self._epic = JiraEpic(source=self.source,key=epic_key)
            if self._epic.load():
                return self._epic
            else:
                return None

        return epic_key

    def get_field(self, field, translate=False):
        if translate:
            parts = field.split(".")
            parts[0] = self.default_fields[parts[0].lower()]
            field = '.'.join(parts)

        if self._issue and self._issue.fields:

            if field not in self._issue:
                parts = field.split(".")
                current = self._issue.fields
                val = None
                for p in parts:
                    if p in current and not isinstance(current[p], dict):
                        val = current[p]
                        break
                    elif p in current:
                        current = current[p]

                return val
            else:
                return self._issue[field]
        else:
            return None

    def prepopulate(self, data):
        """
        Takes a dictionary as returned from the JSON REST API (in JSON form)
        :param data: The issue dictionary
        :return: Returns the issue populated.
        """

        if isinstance(data, Issue):
            self._issue = munchify(data.raw)
        elif isinstance(data, dict) and 'fields' in data:
            self._issue = munchify(data)
        else:
            self._issue = None

        return self._issue


class JiraEpic(JiraIssue):

    def __init__(self, source, **kwargs):
        super(JiraEpic, self).__init__(source, **kwargs)
        self._epic_issues = None

    @property
    def issues(self):
        """
        Gets all the issues contained within the epic as a JiraIssueCollection
        :return:
        """
        if self._epic_issues:
            return self._epic_issues

        self._epic_issues = JiraIssueCollection(source=self.source, input_jql="\"Epic Link\" = %s" % self.key)
        if self._epic_issues.load():
            return self._epic_issues
        else:
            return None


class JiraIssueCollection(JiraObject):
    """
    Represents a collection of issues

    Options:
        - input_jql (Optional) - The JQL used to load the issue set.  Required if jira_issue_list not given.
        - group_id (Optional) - The ID of the group to use for context.  If not provided, then certain features are not
                    available such as "completed_points"
        - input_jira_issue_list (Optional) - A list of json objects as dicts returned from the JIRA REST API. Required
                if jql not given
        - issue_keys (Optional) - A list or comma separated string of issue keys to load
        - paging_start_at (Optional, Default=0) - The issue index to start with
        - paging_max_results (Optional, Default=500) - The maximum number of issues to return
    """

    def __init__(self, source, **kwargs):
        self._issues = None
        super(JiraIssueCollection, self).__init__(source, **kwargs)

    def count(self):
        return len(self._issues)

    def merge(self, collection):
        """
        Creates a new issue collection which contains the set of issues from this and given collection.  Duplicates
        are removed.
        :param collection: JiraIssueCollection to merge
        :return:
        """

        # this will remove duplicates
        issue_set = set(self._issues + collection._issues)

        # convert back to a list.
        merged_list = list(issue_set)

        jic = JiraIssueCollection(source=self.source)
        jic.prepopulate(merged_list)
        return jic

    def dedupe(self):
        issue_set = set(self._issues)
        self._issues = issue_set
        return self

    def __iter__(self):
        return iter(self._issues)

    @property
    def issues(self):
        return self._issues

    @property
    def total_points(self):
        return reduce(lambda x,y: x+y.points, self._issues, 0)

    @property
    def completed_points(self):
        assert(self.option('group_id'))

        points = 0
        context = AugurContext(self.option('group_id'))
        for i in self._issues:
            if context.workflow.is_resolved(status=i.status,resolution=i.resolution):
                points += i.points

        return points

    @property
    def incomplete_points(self):
        assert(self.option('group_id'))

        points = 0
        context = AugurContext(self.option('group_id'))
        for i in self._issues:
            if not context.workflow.is_resolved(status=i.status,resolution=i.resolution):
                points += i.points

        return points

    def prepopulate(self, data):
        """
        Creates a copy of the given JiraIssue objects and populates this with them.  Note that it can take a
        list of JiraIssue objects or another JiraIssueCollection.  In both cases, it makes a copy of the JiraIssues
        within.
        :param data: A list of JiraIssue objects or a JiraIssueCollection
        :return: Returns the given list if successful
        """
        if isinstance(data,list):
            self._issues = copy(data)
            return self._issues
        elif isinstance(data, JiraIssueCollection):
            self._issues = copy(data._issues)

        return self._issues

    def _load(self):

        fields = self.source.default_fields

        jql = None
        if self.option('input_jql'):
            jql = self.option('input_jql')
        elif self.option('issue_keys') is not None:
            if isinstance(self.option('issue_keys'), str):
                keys = self.option('issue_keys').split(',')
            else:
                keys = self.option('issue_keys')

            if len(keys) > 0:
                jql = "key in (%s)"%",".join(keys)
            else:
                # an empty set of keys was given.  This is valid but we can bypass the remaining logic
                self._issues = []
                return True

        if jql:
            self.log_access('search',jql)
            search_results = self.source.jira.search_issues(
                jql,
                startAt=self.option('paging_start_at'),
                maxResults=self.option('paging_max_results', 0),
                validate_query=True,
                fields=list(fields.values()),
                expand="changelog",
                json_result=False)  ## Must set to False to let PyJira manage paging

            if search_results:
                issues = search_results
            else:
                self.logger.error("Unable to load issues from JQL: %s" % jql)
                return False

        elif self.option('input_jira_issue_list'):

            if isinstance(self.option('input_jira_issue_list'), list):
                issues = self.option('input_jira_issue_list')
            else:
                self.logger.error("JiraIssueCollection: Received 'input_jira_issue_list' that was not "
                                  "of type JiraIssueCollection")
                return False
        else:
            self.logger.error("Unable to load issue collection because no source was specified (jql or list)")
            return False

        assert (isinstance(issues, list))
        self._issues = []
        for i in issues:
            issue_ob = JiraIssue(self.source)
            result = issue_ob.prepopulate(i)
            if len(result):
                self._issues.append(issue_ob)

        return True


class JiraReleaseNotes(JiraIssueCollection):

    def __init__(self, source, **kwargs):
        super(JiraReleaseNotes, self).__init__(source, **kwargs)
        self._issues = None
        self._release_notes = None

    def _load(self):
        
        try:
            start, end = self._get_date_range()

        except (TypeError,LookupError) as e:
            self.logger.error(e.message)
            return False

        start_str = start.format("YYYY/MM/DD HH:mm")
        end_str = end.format("YYYY/MM/DD HH:mm")

        context = AugurContext(self.option('group_id'))
        input_jql = "%s AND (status in (\"Resolved\") AND status changed to \"Production\" " \
                    "during ('%s','%s')) order by updated asc" % (context.workflow.get_projects_jql(),
                                                                  start_str, end_str)
        self._set_option('input_jql',input_jql)

        # load the issues that were released during that time
        if not super(JiraReleaseNotes,self)._load():
            return False

        # Collect statistics about those issues and get total points based on parent issues if applicable.
        #   Also, find out which epic it was part of and aggregate the total points released.
        released_tickets = []
        bug_count = 0
        task_story_count = 0
        total_points = 0
        for issue in self._issues:

            if issue.issuetype.lower() in ('bug','defect'):
                bug_count += 1
            else:
                task_story_count += 1

            points = issue.points

            epic = None
            parent = None
            if issue.is_subtask:
                # get the parent ticket instead
                parent = issue.get_parent(only_key=False)
                if parent:
                    epic = parent.get_epic(only_key=False)
                    if not points:
                        points = parent.points
            else:
                epic = issue.get_epic(only_key=False)

            released_tickets.append({
                'issue':issue,
                'parent': parent,
                'epic': epic,
                'points': points,
            })

            total_points += issue.points

        self._release_notes = munchify({
            'start_date': start,
            'end_date': end,
            'total_points': total_points,
            'bug_count': bug_count,
            'feature_count': task_story_count,
            'issues': released_tickets
        })

        return True

    @property
    def start(self):
        return self._release_notes.start_date

    @property
    def end(self):
        return self._release_notes.end_date

    @property
    def total_points(self):
        return self._release_notes.total_points

    @property
    def bug_count(self):
        return self._release_notes.bug_count

    @property
    def feature_count(self):
        return self._release_notes.feature_count

    @property
    def released_issues(self):
        return self._release_notes.issues

    @property
    def release_notes(self):
        return self._release_notes

    def _get_date_range(self):
        """
        Determine the start and end dates to use for the release notes query based on information found
        in the start and end option strings.  These options could be set as strings, Arrow objects or datetime
        objects.  If string, then it can be any one of the formats found in POSSIBLE_DATE_TIME_FORMATS.
        :return: Returns a tuple containing the start and end date as Arrow objects
        """
        start = self.option('start')
        end = self.option('end')
        
        if not start or not end:
            raise LookupError("You must specify a start and end date for releases")
        else:
            if isinstance(start, Arrow):
                calc_start = start
            elif isinstance(start, str):
                calc_start = arrow.get(start, POSSIBLE_DATE_TIME_FORMATS).replace(tzinfo=None)
            elif isinstance(start, datetime.datetime):
                calc_start = arrow.get(start)
            else:
                raise TypeError("Invalid start date type given.  Must be Arrow, string or datetime")

            if isinstance(end, Arrow):
                calc_end = end
            elif isinstance(start, str):
                calc_end = arrow.get(end, POSSIBLE_DATE_TIME_FORMATS).replace(tzinfo=None)
            elif isinstance(end, datetime.datetime):
                calc_end = arrow.get(end)
            else:
                raise TypeError("Invalid start date type given.  Must be Arrow, string or datetime")

        return calc_start, calc_end


from jira.resources import Board, Sprint, GreenHopperResource
from munch import munchify

from augur import db, common
from pony import orm

from augur.integrations.objects.base import JiraObject, InvalidData
from augur.integrations.objects.issue import JiraIssueCollection


class JiraSprint(JiraObject):
    """
    JiraSprint objects can be loaded using a sprint ID for basic information or for more of a report on a sprint,
    it will require both the sprint ID and the board ID.

    Options:
        sprint_id - The ID of the sprint in Jira
        board_id (optional) - The agile board (rapid board) ID in Jira
        include_reports (optional, Default=False) - If True then, when loading from Jira, there will be two calls.
                            One to load the sprint details and one to return the sprint report.
    """

    def __init__(self, source, **kwargs):
        super(JiraSprint, self).__init__(source, **kwargs)
        self._sprint_report = None
        self._sprint = None
        self._epics = None
        self._completed_issues = None
        self._incomplete_issues = None

    def __str__(self):
        if not self._sprint:
            return "<Unpopulated JiraSprint object>"
        else:
            return "%s (start: %s, end: %s)" % (self._sprint.name,
                                                self.start_date.isoformat() if self.start_date else "None",
                                                self.end_date.isoformat() if self.end_date else "None")

    @property
    def start_date(self):
        if not self._sprint:
            self.logger.error("JiraSprint: No sprint has been loaded yet.")
            return False

        if isinstance(self._sprint.startDate, str):
            # convert if necessary
            start_date = self._convert_date_string_to_date_time(self._sprint.startDate)
            if start_date:
                self._sprint.startDate = start_date
            else:
                return None

        return self._sprint.startDate

    @property
    def completed_issues(self):

        if not self._sprint_report:
            self._load_sprint_report()

        if self._completed_issues:
            return self._completed_issues

        ids = [issue.key for issue in self._sprint_report.completedIssues]
        completed_issues = JiraIssueCollection(source=self.source, issue_keys=ids)
        if completed_issues.load():
            self._completed_issues = completed_issues
            return completed_issues
        else:
            return None

    @property
    def incomplete_issues(self):

        if not self._sprint_report:
            self._load_sprint_report()

        if self._incomplete_issues:
            return self._incomplete_issues

        ids = [issue.key for issue in self._sprint_report.issuesNotCompletedInCurrentSprint]
        incomplete_issues = JiraIssueCollection(source=self.source, issue_keys=ids)
        if incomplete_issues.load():
            self._incomplete_issues = incomplete_issues
            return incomplete_issues
        else:
            return None

    @property
    def all_issues(self):
        completed_issues = self.completed_issues
        incomplete_issues = self.incomplete_issues

        return completed_issues.merge(incomplete_issues)

    @property
    def epics(self):
        """
        This gets all the epics that are associated with this collection of issues
        :return:
        """
        if not self._epics:

            if not self._sprint_report:
                self._load_sprint_report()

            all_issues = self.all_issues
            epic_cache = {}

            for issue in all_issues:
                epic_key = issue.get_epic(only_key=True)
                if epic_key not in epic_cache:
                    epic = issue.get_epic(only_key=False)
                    if epic:
                        epic_cache[epic.key] = epic

            self._epics = JiraIssueCollection(source=self.source)
            if not self._epics.prepopulate(list(epic_cache.values())):
                return None

        return self._epics

    @property
    def added_tickets(self):
        if not self._sprint_report:
            self.logger.error("You must provide the completed sprints")
            return 0

        return list(self._sprint_report.issueKeysAddedDuringSprint.keys()) if \
            self._sprint_report.issueKeysAddedDuringSprint else []

    @property
    def punted_tickets(self):
        if not self._sprint_report:
            self._load_sprint_report()

        return list(self._sprint_report.puntedIssues.keys()) if \
            self._sprint_report.puntedIssues else []

    @property
    def punted_points(self):
        if not self._sprint_report:
            self._load_sprint_report()

        return self._sprint_report.puntedIssuesEstimateSum.text if \
            self._sprint_report.puntedIssuesEstimateSum else 0

    @property
    def completed_points(self):
        if not self._sprint_report:
            self._load_sprint_report()

        return self._sprint_report.completedIssuesEstimateSum.value if 'value' in self._sprint_report.completedIssuesEstimateSum else 0

    @property
    def incomplete_points(self):
        if not self._sprint_report:
            self._load_sprint_report()

        return self._sprint_report.issuesNotCompletedEstimateSum.value if 'value' in self._sprint_report.issuesNotCompletedEstimateSum else 0

    @property
    def end_date(self):
        if not self._sprint:
            self.logger.error("JiraSprint: No sprint has been loaded yet.")
            return False

        if isinstance(self._sprint.endDate, str):
            # convert if necessary
            end_date = self._convert_date_string_to_date_time(self._sprint.endDate)
            if end_date:
                self._sprint.endDate = end_date
            else:
                return None

        return self._sprint.endDate

    @property
    def average_point_size(self):
        if not self._sprint_report:
            self._load_sprint_report()

        points = 0.0
        for issue in self._sprint_report.completedIssues:
            if 'estimateStatistic' in issue and 'statFieldValue' in issue.estimateStatistic and \
                    'value' in issue.estimateStatistic.statFieldValue:
                points += float(issue.estimateStatistic.statFieldValue.value)

        if len(self._sprint_report.completedIssues) > 0:
            return points / float(len(self._sprint_report.completedIssues))
        else:
            return 0

    @property
    def completed_date(self):
        if not self._sprint:
            self.logger.error("JiraSprint: No sprint has been loaded yet.")
            return None

        if isinstance(self._sprint.completedDate, str):
            # convert if necessary
            completed_date = self._convert_date_string_to_date_time(self._sprint.completedDate)
            if completed_date:
                self._sprint.completedDate = completed_date
            else:
                return None

        return self._sprint.completedDate

    @property
    def name(self):
        if not self._sprint:
            self.logger.error("JiraSprint: No sprint has been loaded yet")
            return None

        return self._sprint.name

    @property
    def state(self):
        if not self._sprint:
            self.logger.error("JiraSprint: No sprint has been loaded yet")
            return None

        return self._sprint.state

    @property
    def future(self):
        if not self._sprint:
            self.logger.error("JiraSprint: No sprint has been loaded yet")
            return None

        return self._sprint.state.lower() == "future"

    @property
    def active(self):
        if not self._sprint:
            self.logger.error("JiraSprint: No sprint has been loaded yet")
            return None

        return self._sprint.state.lower() == 'active'

    def _load(self):

        if not self.option("sprint_id"):
            self.logger.error("JiraSprint: Sprint ID must be given")
            return False

        if self.option('include_report'):
            if self.option('board_id'):
                if not self._sprint_report:
                    self._load_sprint_report()
            else:
                self.logger.error("JiraSprint: To get a full sprint report you need to provide the board_id option")
                return False
        else:
            if not self._sprint:
                self._load_sprint()

        return True

    def _proxy_sprint_report(self, board_id, sprint_id):
        """Returns the full sprint report details as is from the API"""
        return

    def _load_sprint_report(self):
        sprint_id = self.option("sprint_id")
        board_id = self.option("board_id")

        self.log_access('sprint-full', board_id, sprint_id)
        spr = self.source.jira._get_json(
            'rapid/charts/sprintreport?rapidViewId=%s&sprintId=%s' % (board_id, sprint_id),
            base=self.source.jira.AGILE_BASE_URL,
            params={'agile_rest_path': GreenHopperResource.GREENHOPPER_REST_PATH})

        sprint_report = munchify(spr)

        if sprint_report:

            # only grab the contents area which contains the meatiest bits.
            self._sprint_report = munchify(sprint_report.contents)

            # the sprint report is a superset of the sprint object.  The sprint key in the sprint report object
            #   contains the same keys plus a few more so we can reuse it for the sprint object so that both
            #   can be used without having to do a load twice.
            self._sprint = munchify(sprint_report.sprint)

            return True
        else:
            self.logger.error("JiraSprint: Unable to load sprint report data from Jira")
            return False

    def _load_sprint(self):

        sprint_id = self.option("sprint_id")
        self.log_access('sprint-brief', sprint_id)
        s = self.source.jira.sprint(sprint_id)
        if s:
            self._sprint = munchify(s)
        else:
            self._sprint = None
            self.logger.error("JiraSprint: Unable to find the sprint with this id: %s" % str(sprint_id))
            return False

    def prepopulate(self, data):
        if 'contents' in data:
            self._sprint_report = munchify(data['contents'])
            self._sprint = munchify(data['sprint'])
            return True

        elif 'state' in data and 'name' in data:
            self._sprint = munchify(data)
            if self.option('include_reports'):
                if not self.option('board_id'):
                    self.logger.error("You cannot load reports within a JiraSprint object without a board ID given")
                    return False

                # do not return False if this fails.
                self._load_sprint_report()
            return True
        else:
            self.logger.error("JiraSprint: Could not prepopulate because the given data does not match expected values")
            return False

    @property
    def sprint_id(self):
        return self._sprint_id

    @sprint_id.setter
    def sprint_id(self, val):
        self._sprint_id = val

    @property
    def details(self):
        return self._sprint

    @property
    def report(self):
        if not self._sprint_report:
            self._load_sprint_report()

        return self._sprint_report


class JiraSprintCollection(JiraObject):
    """
    Represents a collection of issues

    Options:
        - board (Optional) - This can either be a board ID or a JiraBoard object.
        - sprints (Optional) - The JIRA sprint objects as returned from the Jira REST API.
        - max_sprints (optional, default=None) - If given this is the maximum number of sprints to load keeping
                            in mind that they are loaded in reverse chronological order.
        - include_reports (optional, default=False) - If True, this will set make extra calls per
                            sprint to retrieve detailed sprint reports.
        - team_name - If given, this will restrict the sprints to those which have the given team name in the title.
        - states (optional, default="['closed','active']) - Which sprints to include based on state.  By default we
                            exclude future sprints.
    """

    def __init__(self, source, **kwargs):
        super(JiraSprintCollection, self).__init__(source, **kwargs)
        self._sprints = None
        if not self.option('states'):
            self._options.states = ['closed', 'active']

    def __iter__(self):
        return iter(self._sprints)

    def __str__(self):
        return "Sprint Collection:\n\t%s" % format("\n\t".join([str(s) for s in self._sprints]))

    @property
    def board_id(self):
        board = self.option('board')
        if board and isinstance(board, JiraBoard):
            return board.id
        elif board:
            return int(board)

        return None

    def _load(self):
        if not self.option('sprints') and not self.option('board'):
            self.logger.error("JiraSprintCollection: No sprint data source provided")
            return False

        if not self._sprints:

            if self.option('board'):
                board = self.option('board')
                if isinstance(board, JiraBoard):
                    board_id = board.id
                else:
                    board_id = board

                # we do it this way because this returns a paginated object that automatically
                #   makes additional calls when there are more than fit on a single page.
                self.log_access('sprints', board_id)
                sprints = self.source.jira.sprints(board_id, maxResults=0, state=','.join(self.option('states')))

            elif self.option('sprints'):
                sprints = self.option('sprints')
            else:
                # should never get here because initial check should validate
                #   required input.
                self.logger.error("You must provide a non empty set of sprints or a board to load a sprint collection")
                return False

            sprints.reverse()
            self._sprints = []
            for s in sprints:
                if self.option('max_sprints') and len(self._sprints) > int(self.option('max_sprints')):
                    # there's no need to look at any more sprints if we've
                    #   hit the maximum requested.
                    break

                if isinstance(s, dict):
                    jira_sprint_json = s
                elif isinstance(s, Sprint):
                    jira_sprint_json = s.raw
                else:
                    self.logger.error("Unrecognized sprint object found. Skipping...")
                    continue

                sprint_ob = JiraSprint(source=self.source, sprint_id=jira_sprint_json['id'],
                                       board_id=self.board_id,
                                       include_reports=self.option('include_reports'))

                continue_adding = True

                # filter on sprints with names that match the team (if given)
                if self.option('team_name'):
                    continue_adding = common.find_team_name_in_string(self.option('team_name'),
                                                                      jira_sprint_json['name'])

                if jira_sprint_json['state'].lower() not in self.option('states'):
                    continue_adding = False

                if continue_adding:
                    sprint_ob.prepopulate(jira_sprint_json)
                    self._sprints.append(sprint_ob)

            # order them from most recent to oldest by default.

        return self._sprints


class JiraBoard(JiraObject):
    """
    Provides access to a Jira Agile Board.
    Options:
        - team_id OR board_id OR both- If team ID given then board id will be taken from the database (if it exists). If
                                board ID is given but not team id then
        - max_sprints (optional, default=None) - If given this is the maximum number of sprints to load keeping
                            in mind that they are loaded in reverse chronological order.
        - restrict_sprints_with_team_name (optional, default=False) - If a team is given and if this is set to
                                True then only sprints that match the team name will be used.
        - include_sprint_reports (optional, default=False) - If this is specified loading the board will
                                take much longer as a separate call is made to retrieve the report
                                for each sprint.

    """

    def __init__(self, source, **kwargs):
        super(JiraBoard, self).__init__(source, **kwargs)
        self._team = None
        self._db_board = None
        self._jira_board = None
        self._sprints = None
        self._backlog_issues = None

    @property
    def id(self):
        if self._jira_board:
            return self._jira_board.id
        else:
            self.logger.error("JiraBoard: A board object has not yet been loaded")
            return None

    def get_backlog(self):
        """
        Returns a JiraIssueCollection of all the issues stored in the backlog.
        :return: Returns a JiraIssueCollection populated with backlog issues.
        """
        if not self._db_board:
            self.logger.error("JiraBoard: There is no board ID specified for this yet")
            return None

        if not self._backlog_issues:
            self.log_access('sprint-backlog', self._db_board.jira_id)
            result = self.source.jira.get_backlog_issues(self._db_board.jira_id, json_result=True)
            if result:
                try:
                    self._backlog_issues = JiraIssueCollection(self.source, input_jira_issue_list=result['issues'])
                    self._backlog_issues.load()
                except InvalidData as e:
                    self.logger.warn("Unable to retrieve backlog.  %s" % e.message)
                    return None
            else:
                self._backlog_issues = None

        return self._backlog_issues

    def get_sprint(self, sprint_id, include_reports=False):
        """
        Gets the sprint with the given id
        :param sprint_id: The ID of the sprint to load
        :param include_reports: True to load the detailed sprint report in addition to the normal sprint data.
        :return:
        """
        # if we haven't pulled all the sprints then just make a request.
        sprint = JiraSprint(self.source, sprint_id=sprint_id, include_reports=include_reports)
        if sprint.load():
            return sprint
        else:
            return None

    def get_sprints(self):
        """
        Gets the sprints on the agile board in order from newest to oldest.  Note that this will not include
        future sprints.
        :return: Returns a JiraSprintCollection
        """
        if not self._sprints:

            if self.option('restrict_sprints_with_team_name'):
                team_name = self._team.name if self._team else ""
            else:
                team_name = None

            self._sprints = JiraSprintCollection(self.source, board=self,
                                                 include_reports=self.option('include_sprint_reports'),
                                                 max_sprints=self.option('max_sprints'),
                                                 team_name=team_name,
                                                 include_future=False)
            self._sprints.load()

        return self._sprints

    def get_most_recent_active_sprint(self):
        sprints = self.get_sprints()
        for s in sprints:
            if s.active:
                return s

        return None

    def get_most_recent_closed_sprint(self):
        return self.get_past_sprint(number_of_sprints_in_the_past=1)

    def get_past_sprint(self, number_of_sprints_in_the_past):
        sprints = self.get_sprints()
        i = 1
        for s in sprints:
            # the first inactive one it finds it returns since we assume that they are orded from newest to oldest
            #   (see get_sprints)
            if s.state.lower() == 'closed':
                if i == number_of_sprints_in_the_past:
                    return s
                else:
                    i += 1

        return None

    def _load(self):
        if not self.option('board_id') and not self.option('team_id'):
            self.logger.error("Board ID or Team ID must be given")
            return None

        if self._load_db_data():
            return self._load_jira_data()

        return False

    def _load_db_data(self):
        team_id = self.option('team_id')
        if team_id:
            team = db.Team[team_id]
            if team:
                board = team.agile_board
            else:
                self.logger.error("Unable to find team with ID %d" % team_id)
                return False
        else:
            board_id = self.option('board_id')
            board = orm.select(b for b in db.AgileBoard if b.jira_id == board_id).first()
            if board:
                team = board.team
            else:
                self.logger.error("Unable to find board with ID %d" % board_id)
                return False

        self._team = team
        self._db_board = board

        return self._team is not None and self._db_board is not None

    def _load_jira_data(self):

        board_id = self.option('board_id')
        if not board_id:
            board_id = self._db_board.jira_id if self._db_board else None

        if board_id:

            self.log_access('board', board_id)

            b = Board(self.source.jira._options, self.source.jira._session)
            b._resource = 'board/{0}'  # seems to be a bug with the find method
            b.find(board_id)

            self._jira_board = munchify(b.raw)

            if not b:
                self.logger.error("Could not load the board information from Jira")
                return False
        else:
            self.logger.error("Could not get a board ID to load the board with")
            return False

        return True

    @property
    def board(self):
        return self._db_board

    @property
    def team(self):
        return self._team



"""
AUGUR API

The Augur API provides easy access to some of the most benefecial data 
that can be collected by the library.  While not exhaustive it is meant
to conveniently retrieve the most commonly needed data.

Rather than instantiating data classes or integrations directly, the API 
provides a layer of abstraction in the form of functions.  Data that is 
returned is almost always in the form of a dictionary that can be easily 
converted into JSON as necessary.

Most functions within the api will take an AugurContext object which
defines the constraints within which the data should be retrieved
and interpreted. 

See the Augur readme for some information about how data is organized,
the meaning of groups, workflows, etc.

"""

import datetime
import logging

import copy

from munch import munchify
from pony import orm
from pony.orm import select, delete

from augur import settings
from augur import db
from augur.db import EventLog
from augur.integrations.objects import JiraBoard, BoardMetrics, JiraIssue

CACHE = dict()

__jira = None
__github = None
__context = None

api_logger = logging.getLogger("augurapi")


def set_default_context(context):
    """
    Sets the default context for the API
    :param context: The context object to use by default.
    :return:
    """
    global __context
    __context = context


def get_default_context():
    """
    Returns the default context for the API
    :return: Returns the default context object (AugurContext)
    """
    global __context
    return __context


def get_github():
    """
    Returns the global github integration object
    :return: Returns the global AugurGithub instance
    """
    from augur.integrations.augurgithub import AugurGithub

    global __github
    if not __github:
        __github = AugurGithub()
    return __github


def get_jira():
    """
    Returns the global jira integration object
    :return: Returns the global AugurJira instance
    """
    from augur.integrations.augurjira import AugurJira

    global __jira
    if not __jira:
        __jira = AugurJira()
    return __jira


def get_jira_url():
    return settings.main.integrations.jira.instance


def get_workflow(workflow_id):
    """
    Gets a workflow by ID
    :param workflow_id: The ID of the workflow to retrieve
    :return: Returns a Group object or None if not found
    """
    return db.Workflow[workflow_id]


def get_workflows():
    """
    Get a list of all workflows
    :return: Returns a Group object or None if not found
    """
    return orm.select(w for w in db.Workflow)


def get_group(group_id):
    """
    Gets a group by id
    :param group_id: The ID or name of the group to retrieve
    :return: Returns a Group object or None if not found
    """
    return db.Group[group_id]


def get_groups():
    """
    Gets all groups
    :return: Returns a Group object or None if not found
    """
    return orm.select(g for g in db.Group)[:]


def get_vendors():
    """
    Gets all the vendors stored in the database
    :return: Returns a list of vendors
    """
    return orm.select(v for v in db.Vendor)[:]


def get_issue_details(key):
    """
    Return details about an issue based on an issue key.  This does not pull from any cache.
    :param key: The key of the issue to retrieve
    :return: The issue dict
    """

    issue = JiraIssue(source=get_jira(), key=key)
    if issue.load():
        return issue.issue
    else:
        return None


def get_all_staff(context):
    """
    :param context: The context to use in determining which staff to return (None returns all)
    """
    if not context:
        return orm.select(s for s in db.Staff).order_by(lambda x: x.last_name)[:]
    else:
        context_staff = []
        valid_team_ids = [t.id for t in context.group.teams]
        staff = orm.select(s for s in db.Staff if s.teams)[:]
        for s in staff:
            for t in s.teams:
                if t.id in valid_team_ids:
                    context_staff.append(s)
        return context_staff


def get_staff_member_by_field(first_name=None, last_name=None, email=None, username=None):
    """
    Search for a staff member by fields other than ID.  This will try to match on email first (if given)
    then will try to find it based on first,last name.  First and last name must be given to do the search.
    :param username: One of the stored usernames (e.g. jira or github)
    :param first_name: The first name of the user (if given, last name must also be given)
    :param last_name: The last name of the user (if given, first name must also be given.
    :param email: The email of the user to find.
    :return: Returns a db.Staff object if found, otherwise None.
    """

    def search_by_username():
        try:
            # try searching by first, last name
            return orm.get(s for s in db.Staff if s.github_username == username or
                           s.jira_username == username)
        except (orm.MultipleObjectsFoundError, orm.ObjectNotFound):
            return False

    def search_by_name():
        try:
            # try searching by first, last name
            if first_name and last_name:
                return orm.get(s for s in db.Staff if s.first_name.lower() == first_name.lower() and
                               s.last_name.lower() == last_name.lower())
            elif first_name:
                return orm.get(s for s in db.Staff if s.first_name.lower() == first_name.lower())
            elif last_name:
                return orm.get(s for s in db.Staff if s.last_name.lower() == last_name.lower())

        except (orm.MultipleObjectsFoundError, orm.ObjectNotFound):
            return False

    def search_by_email():
        try:
            # try searching by first, last name
            return orm.get(s for s in db.Staff if s.email.lower() == email.lower())
        except (orm.ObjectNotFound, orm.MultipleObjectsFoundError):
            return False

    result = None
    if username:
        result = search_by_username()

    if not result and email:
        result = search_by_email()

    if not result and (first_name or last_name):
        result = search_by_name()

    return result


def get_team_by_id(team_id):
    """
    Returns a team object keyed on the name of the team.  If there is more than one team with the same name it will
    always returns only 1.  There is no guarantee which one, though.
    :param team_id: The team id to search for
    :return: A Team object.
    """
    return db.Team[team_id]


def get_teams(context=None):
    """
    Retrieves a list of team objects containing all the known teams in e-comm
    :return: An array of Team objects. 
    """
    if not context:
        return select(t for t in db.Team).order_by(lambda x: x.name)[:]
    else:
        return context.group.teams.order_by(lambda x: x.name)


def get_products():
    """
    Retrieves a list of product objects containing all the known product groups in e-comm
    :return: An array of Product objects. 
    """
    return select(p for p in db.Product)


def memory_cache_data(data, key):
    """
    Cache data in memory
    :param data: The data to cache
    :param key: The key to store it under
    :return: Returns the data given in <data>
    """
    global CACHE
    CACHE[key] = copy.deepcopy(data)
    return CACHE[key]


def get_memory_cached_data(key):
    """
    Retrieves data stored in memory under <key>
    :param key: The key to look for in the in-memory cache
    :return: Returns the data or None if not found
    """
    global CACHE
    if key in CACHE:
        return CACHE[key]
    else:
        return None


def get_board_metrics(board_id, context):
    """
    Retrieves information about the backlog  for the given board.
    :param board_id:  The ID of the board to get the backlog from
    :param context: The context to help define how to interpret the data retrieved
    :return: Returns a dict with information about the backlog
    """
    board = JiraBoard(get_jira(), board_id=board_id)
    if board.load():
        metrics = BoardMetrics(context, board)
        return munchify({
            'sprints': metrics.historic_sprint_analysis(),
            'backlog': metrics.backlog_analysis()
        })
    else:
        return None


def get_issue_field_from_custom_name(name):
    """
    Returns the true field name of a jira field based on its friendly name
    :param name: The friendly name of the field
    :return: A string with the true name of a field.
    """
    return get_jira().get_field_by_name(name)


def add_staff(staff_properties):
    """
    Staff properties must include the following:
        * first_name (unicode)
        * last_name (unicode)
        * company (unicode)
        * type (unicode - one of STAFF_TYPES)
        * role (unicode - one of STAFF_ROLES)
        * email (unicode)
        * rate (if consultant)
        * jira_username
        * team (db.Team model)
    :param staff_properties:
    :return: Returns a db.Staff object.
    """
    db.Staff(**staff_properties)
    orm.commit()


def update_staff(staff_id, staff_properties):
    """
    Staff properties can be any subset of the properties that are
    part of the Staff model.
    :param staff_id: The ID of the staff object to update
    :param staff_properties: All or some of the properties in the staff model
    :return:
    """
    s = db.Staff[staff_id]
    if s:
        s.set(**staff_properties)
        orm.commit()
        return s

    return None


def add_team(team_properties):
    """
    Team properties must include the following:
        * name (unicode)
    :param team_properties: At a minimum, contains the required fields necessary to create a team.
    :return: Returns a db.Team object.
    """
    db.Team(**team_properties)
    orm.commit()


def update_team(team_id, team_properties):
    """
    Staff properties can be any subset of the properties that are
    part of the Team model.
    :param team_id: The ID of the team object to update
    :param team_properties: All or some of the properties in the team model
    :return:
    """
    t = db.Team[team_id]
    if t:
        t.set(**team_properties)
        orm.commit()
        return t

    return None


def clear_event_data(days_to_keep=30):
    """
    Remove all logs older than days_to_keep days
    :param days_to_keep: The number of days to keep and remove all other log entries.
    :return: Returns the number of rows deleted
    """
    return delete(
        e for e in EventLog if e.event_time < (datetime.datetime.now() - datetime.timedelta(days=days_to_keep)))


def log_event_data(event_type, event_data):
    """
    Stores a log entry in the database
    :param event_type:
    :param event_data:
    :return:
    """
    event_time = datetime.datetime.now()
    el = EventLog(event_time=event_time, event_type=event_type, event_data=event_data)
    return el is not None



from pony import orm
from pony.orm import sql_debug, Json
import datetime

NOTIFY_TYPES = ["none", "email", "slack"]
TOOL_ISSUE_STATUS_TYPES = ["open", "in progress", "done"]
STATUS = ["Unknown", "Active", "Inactive", "Pending"]
ROLES = ["Unknown", "Developer", "SDET", "Technical Manager", "Product Manager", "Director", "Lead Developer",
         "Engagement Manager", "Project Manager",
         "Business Analyst", "QA", "Technical Manager", "Product Owner", "Vice President"]
TOOL_ISSUE_RESOLUTION_TYPES = ["positive", "negative"]
TOOL_ISSUE_TYPE_TYPES = ["story", "task", "bug", "question"]
BUILD_TYPES = ['all', 'upstream', 'myteam', 'otherteams']
STAFF_TYPES = ["FTE", "Consultant"]

db = orm.Database()
__is_bound = False


class ToolIssueResolution(db.Entity):
    """
    Represents an issue type within a workflow tool.  For example, Jira.
    """
    id = orm.PrimaryKey(int, auto=True)
    tool_issue_resolution_name = orm.Required(str)
    tool_issue_resolution_type = orm.Required(str, py_check=lambda v: v in TOOL_ISSUE_RESOLUTION_TYPES)
    workflows = orm.Set('Workflow', reverse="resolutions")


class ToolIssueStatus(db.Entity):
    """
    Represents an issue type within a workflow tool.  For example, Jira.
    """
    id = orm.PrimaryKey(int, auto=True)
    tool_issue_status_name = orm.Required(str)
    tool_issue_status_type = orm.Required(str, py_check=lambda v: v in TOOL_ISSUE_STATUS_TYPES)
    workflows = orm.Set('Workflow', reverse="statuses")


class ToolIssueType(db.Entity):
    """
    Represents an issue type within a workflow tool.  For example, Jira.
    """
    id = orm.PrimaryKey(int, auto=True)
    tool_issue_type_name = orm.Required(str)
    tool_issue_type_type = orm.Required(str, py_check=lambda v: v in TOOL_ISSUE_TYPE_TYPES)
    workflow_defect_project_filters = orm.Set('WorkflowDefectProjectFilter', reverse="issue_types")
    workflows = orm.Set('Workflow', reverse="issue_types")


class EventLog(db.Entity):
    """
    Represents an event log entry
    """
    id = orm.PrimaryKey(int, auto=True)
    event_time = orm.Required(datetime.datetime)
    event_type = orm.Required(str)
    event_data = orm.Optional(Json)


class Vendor(db.Entity):
    """
    Represents a third-party vendor who is either providing services or products to the group
    """
    id = orm.PrimaryKey(int, auto=True)
    name = orm.Required(str)
    engagement_contact_first_name = orm.Optional(str)
    engagement_contact_last_name = orm.Optional(str)
    engagement_contact_email = orm.Optional(str)
    billing_contact_first_name = orm.Optional(str)
    billing_contact_last_name = orm.Optional(str)
    billing_contact_email = orm.Optional(str)
    tempo_id = orm.Optional(int)
    consultants = orm.Set('Staff', reverse='vendor')

    def get_engagement_contact_full_name(self):
        if self.engagement_contact_first_name and self.engagement_contact_last_name:
            return "%s %s" % (self.engagement_contact_first_name, self.engagement_contact_last_name)
        elif self.engagement_contact_last_name:
            return self.engagement_contact_last_name
        elif self.engagement_contact_first_name:
            return self.engagement_contact_first_name
        else:
            return "None Given"

    def get_billing_contact_full_name(self):
        if self.billing_contact_first_name and self.billing_contact_last_name:
            return "%s %s" % (self.billing_contact_first_name, self.billing_contact_last_name)
        elif self.billing_contact_last_name:
            return self.engagement_contact_last_name
        elif self.billing_contact_first_name:
            return self.engagement_contact_first_name
        else:
            return "None Given"


class Product(db.Entity):
    """
    Represents a single member of a team.  A team member can be on multiple teams and a team can have multiple
     team members.  The staff object is used to store information like hourly rate (when a consultant), usernames
     in various integrations along with start date.
    """
    id = orm.PrimaryKey(int, auto=True)
    name = orm.Required(str)
    key = orm.Required(str, unique=True)
    teams = orm.Set('Team', reverse='product')
    groups = orm.Set('Group', reverse='products')


class ToolProject(db.Entity):
    id = orm.PrimaryKey(int, auto=True)
    tool_project_key = orm.Required(str)
    workflows = orm.Set('Workflow', reverse="projects")


class ToolProjectCategory(db.Entity):
    """
    Represents the category of a project as represented by the associated tools.  Not all tools organize projects
    into categories
    """
    id = orm.PrimaryKey(int, auto=True)
    tool_category_name = orm.Required(str)
    workflows = orm.Set('Workflow', reverse="categories")


class Staff(db.Entity):
    """
    Represents a single member of a team.  A team member can be on multiple teams and a team can have multiple
     team members.  The staff object is used to store information like hourly rate (when a consultant), usernames
     in various integrations along with start date.
    """
    first_name = orm.Required(str)
    last_name = orm.Required(str)
    company = orm.Optional(str)
    avatar_url = orm.Optional(str)
    role = orm.Required(str, py_check=lambda v: v in ROLES)
    email = orm.Required(str)
    rate = orm.Required(float)
    start_date = orm.Required(datetime.date)
    type = orm.Required(str, py_check=lambda v: v in STAFF_TYPES, default="FTE")
    jira_username = orm.Required(str)
    slack_id = orm.Optional(str)
    github_username = orm.Optional(str)
    status = orm.Required(str, py_check=lambda v: v in STATUS)
    teams = orm.Set('Team', reverse="members")
    base_daily_cost = orm.Optional(float)
    base_weekly_cost = orm.Optional(float)
    base_annual_cost = orm.Optional(float)
    vendor = orm.Optional(Vendor, reverse='consultants')
    notification = orm.Optional("Notifications", reverse="staff")

    def get_company(self):
        if self.vendor:
            return self.vendor.name
        else:
            return "None Given"

    def is_in_team(self, team_name):
        for t in self.teams:
            if team_name.lower() in t.name.lower():
                return True
        return False

    def get_fullname(self):
        if self.first_name and self.last_name:
            return "%s %s" % (self.first_name, self.last_name)
        else:
            return self.first_name if self.first_name else self.last_name or "None Given"

    def before_insert(self):
        self.calculate_costs()

    def before_update(self):
        self.calculate_costs()

    def calculate_costs(self):

        # Calculate the cost of the employee post import
        if self.status.lower() == "active":
            self.base_daily_cost = self.rate * 8
            self.base_weekly_cost = self.base_daily_cost * 5
            self.base_annual_cost = self.base_weekly_cost * 50  # assume two weeks of vacation
        else:
            self.base_daily_cost = 0.0
            self.base_weekly_cost = 0.0
            self.base_annual_cost = 0.0


class AgileBoard(db.Entity):
    id = orm.PrimaryKey(int, auto=True)
    jira_id = orm.Optional(int)
    team = orm.Optional('Team', reverse='agile_board')


class Team(db.Entity):
    """
    Represents a team in the organization.  Teams contain staff members and are associated with agile boards
    and products most of the time.
    """
    id = orm.PrimaryKey(int, auto=True)
    name = orm.Required(str)
    members = orm.Set(Staff, reverse='teams')
    agile_board = orm.Optional(AgileBoard, reverse='team', sql_default=0)
    product = orm.Optional(Product, reverse='teams', sql_default=0)
    groups = orm.Set('Group', reverse="teams")
    notification = orm.Optional("Notifications", reverse="team")

    def get_agile_board_jira_id(self):
        return self.agile_board.jira_id if self.agile_board else 0

    def get_groups_as_string(self):
        names = []
        for g in self.groups:
            names.append(g.name)
        return ', '.join(names) if len(names) > 0 else "None"


class Notifications(db.Entity):
    """
    Represents a set of notification preferences that can be associated with a staff member or team
    """
    id = orm.PrimaryKey(int, auto=True)
    build = orm.Required(str, default="email,slack")
    deploy = orm.Required(str, default="email,slack")
    staff = orm.Optional(Staff)
    team = orm.Optional(Team)
    build_types = orm.Required(str, default="all")

    def should_send_build_x(self, ntype):
        ntypes = str(self.build).split(",")
        return ntype in ntypes

    def should_send_deploy_x(self, ntype):
        ntypes = str(self.deploy).split(",")
        return ntype in ntypes

    def should_send_build_email(self):
        return self.should_send_build_x("email")

    def should_send_build_slack(self):
        return self.should_send_build_x("slack")

    def should_send_deploy_email(self):
        return self.should_send_deploy_x("email")

    def should_send_deploy_slack(self):
        return self.should_send_deploy_x("slack")


class WorkflowDefectProjectFilter(db.Entity):
    id = orm.PrimaryKey(int, auto=True)
    project_key = orm.Required(str)
    issue_types = orm.Set(ToolIssueType, reverse='workflow_defect_project_filters')
    workflows = orm.Set('Workflow', reverse="defect_projects")

    def get_issue_types_as_string_list(self):
        types = []
        for it in self.issue_types:
            types.append(it.tool_issue_type_name)

        return types


class Workflow(db.Entity):
    id = orm.PrimaryKey(int, auto=True)
    name = orm.Required(str)
    statuses = orm.Set(ToolIssueStatus, reverse="workflows")
    resolutions = orm.Set(ToolIssueResolution, reverse="workflows")
    projects = orm.Set(ToolProject, reverse="workflows")
    categories = orm.Set(ToolProjectCategory, reverse="workflows")
    issue_types = orm.Set(ToolIssueType, reverse="workflows")
    defect_projects = orm.Set(WorkflowDefectProjectFilter, reverse="workflows")
    groups = orm.Set('Group', reverse="workflow")

    def get_defect_projects(self):
        """
        Gets all the projects that are considered defect projects for this workflow
        :return: Returns a list of dicts that look like this:
            {
                [
                    key: <string>,
                    issue_types: [<string>, <string>, ...]
                ]
            }
        """
        defect_projects = [p for p in self.defect_projects]

        if defect_projects:
            # if this workflow specifies projects with keys and the caller
            #   just wants keys returned then we can just return this list.
            return [dict({"key": df.project_key,
                          "issue_types": df.get_issue_types_as_string_list()})
                    for df in defect_projects]
        else:
            return []

    def is_defect_ticket(self, project_key, issue_type):
        if not project_key or not issue_type:
            return False

        for df in self.defect_projects:
            if project_key.lower() == df.project_key.lower():
                for df_type in df.issue_types:
                    if df_type.tool_issue_type_name.lower() == issue_type.lower():
                        return True
        return False

    def status_ob_from_string(self, status_name):
        try:
            return [x for x in self.statuses if x.tool_issue_status_name.lower() == status_name.lower()].pop()
        except IndexError:
            return None

    def resolution_ob_from_string(self, res_name):
        try:
            return [x for x in self.resolutions if x.tool_issue_resolution_name.lower() == res_name.lower()].pop()
        except IndexError:
            return None

    def is_resolved(self, status, resolution):
        """
        Determines if the given status and resolution indicates a completed ticket
        :param status:
        :type status: ToolIssueStatus
        :param resolution:
        :type resolution: ToolIssueResolution
        :return: Returns False if status or resolution could not be found in this workflow
        """
        status_ob = self.status_ob_from_string(status)
        res_ob = self.resolution_ob_from_string(resolution)
        if not status_ob:
            return False

        if status_ob.tool_issue_status_type.lower() == "done":
            if res_ob:
                if res_ob.tool_issue_resolution_type.lower() == "positive":
                    return True
            else:
                # if the resolution isn't set then we will assume
                return True

        return False

    def is_abandoned(self, status, resolution):
        """
        Determines if the given status and resolution indicates an abandoned ticket. An abandoned
        ticket is a "done" ticket that has a "negative" resolution.
        :param status:
        :type status: ToolIssueStatus
        :param resolution:
        :type resolution: ToolIssueResolution
        :return:
        """
        status_ob = self.status_ob_from_string(status)
        res_ob = self.resolution_ob_from_string(resolution)
        if not status_ob:
            return False

        if status_ob.tool_issue_status_type.lower() == "done":
            if res_ob:
                if res_ob.tool_issue_resolution_type.lower() == "negative":
                    return True

        return False

    def positive_resolutions(self):
        return [x for x in self.resolutions if x.tool_issue_resolution_type.lower() == "positive"]

    def negative_resolutions(self):
        return [x for x in self.resolutions if x.tool_issue_resolution_type.lower() == "negative"]

    def open_statuses(self):
        """
        Returns a list of all the statuses that are considered to be "open".
        :return: A list of ToolIssueStatus objects.
        """
        return [x for x in self.statuses if x.tool_issue_status_type.lower() == "open"]

    def done_statuses(self):
        """
        Returns a list of all the statuses that are considered to be "done".
        :return: A list of ToolIssueStatus objects.
        """
        return [x for x in self.statuses if x.tool_issue_status_type.lower() == "done"]

    def in_progress_statuses(self):
        """
        Returns all statuses that are considered "in progress" according to this workflow
        :return: A list of ToolIssueStatus objects
        """
        return [x for x in self.statuses if x.tool_issue_status_type.lower() == "in progress"]

    def dev_issue_types(self):
        """
        Returns all issue types that are considered development tickets as opposed to bug related tickets.
        :return: A list of ToolIssueType objects
        """
        return [x for x in self.issue_types if x.tool_issue_type_type.lower() in ["story", "task"]]

    def is_in_progress(self, status):
        """
        Returns True if the given status string is considered an "in progress" status.
        :param status: The status string
        :type status: str
        :return: Returns boolean
        :rtype: bool
        """
        for s in self.statuses:
            if s.tool_issue_status_name.lower() == status.lower():
                return s.tool_issue_status_type == "in progress"
        return False

    def get_project_keys(self):
        return self.get_projects(key_only=True)

    def get_projects_by_category(self, category):
        """
        Gets all projects with the given category
        :param category:
        :return:
        """
        from augur import api
        cache_key = "projects_%s" % category
        projects = api.get_memory_cached_data(cache_key)
        if not projects:
            projects = api.get_jira().jira.get_projects_with_category(category)
            api.memory_cache_data({'data': projects}, cache_key)
        else:
            projects = projects[0]['data']

        return projects

    def get_projects(self, key_only=False):
        from augur.api import get_jira
        project_keys = [p.tool_project_key for p in self.projects]
        projects = []

        if project_keys:
            if key_only:
                # if this workflow specifies projects with keys and the caller
                #   just wants keys returned then we can just return this list.
                return project_keys
            else:
                # Call into jira to get the list of projects with these keys
                projects = get_jira().jira.get_projects_with_key(project_keys)
        elif self.categories:
            # Call into jira to get the list of projects with the given categories.
            for c in self.categories:
                projects.extend(self.get_projects_by_category(c.tool_category_name))

        if key_only:
            return [p['key'].upper() for p in projects]
        else:
            return projects

    def get_projects_jql(self):
        """
        Generates the JQL for the portion of the expression that limits the projects to this workspace only.
        :return: Returns a string with the projects JQL
        """
        keys = []
        for p in self.projects:
            keys.append(p.tool_project_key)

        categories = []
        for c in self.categories:
            categories.append(c.tool_category_name)

        jql_projects = ""
        jql_categories = ""
        if len(keys) > 0:
            jql_projects = "project in (%s)" % ",".join(keys)

        if len(categories) > 0:
            jql_categories = "category in ('%s')" % "','".join(categories)

        if jql_projects and jql_categories:
            return "((%s) OR (%s))" % (jql_projects, jql_categories)
        else:
            return jql_projects if jql_projects else jql_categories

    def get_positive_resolution_jql(self):
        """
        Generates the JQL for the portion of the expression that limits the issues to those that have positive
        resolutions (which is defined by the statuses and resolutions defined in this workflow)
        :return: Returns a JQL string
        """
        done_statuses = [d.tool_issue_status_name for d in self.done_statuses()]
        done_resolutions = [d.tool_issue_resolution_name for d in self.positive_resolutions()]

        return "(status in (\"%s\") and resolution in (\"%s\"))" % \
               ('","'.join(done_statuses), '","'.join(done_resolutions))

    def as_json(self):
        return {
            "statuses": {
                "open": [s.tool_issue_status_name for s in self.open_statuses()],
                "in progress": [s.tool_issue_status_name for s in self.in_progress_statuses()],
                "done": [s.tool_issue_status_name for s in self.done_statuses()]
            },
            "projects": self.get_project_keys(),
            "resolutions": {
                "positive": [pr.tool_issue_resolution_name for pr in self.positive_resolutions()],
                "negative": [pr.tool_issue_resolution_name for pr in self.negative_resolutions()],
            }
        }


class Group(db.Entity):
    id = orm.PrimaryKey(int, auto=True)
    name = orm.Required(str)
    workflow = orm.Optional(Workflow, reverse="groups")
    products = orm.Set(Product, reverse="groups")
    teams = orm.Set(Team, reverse="groups")


def init_db():
    global __is_bound

    if not __is_bound:

        from augur import settings

        if settings.main.project.debug:
            sql_debug(True)

        if settings.main.datastores.main.type == "sqlite":
            dbtype = settings.main.datastores.main.type
            dbtarget = settings.main.datastores.main.sqlite.path
            db.bind('sqlite', filename=dbtarget, create_db=True)
            __is_bound = True
        elif settings.main.datastores.main.type == "postgres":
            pg_settings = settings.main.datastores.main.postgres
            dbtype = settings.main.datastores.main.type
            dbtarget = pg_settings.host + "/" + pg_settings.dbname + ":" + pg_settings.port
            db.bind('postgres', user=pg_settings.username, password=pg_settings.password,
                    host=pg_settings.host, database=pg_settings.dbname, port=pg_settings.port)
            __is_bound = True
        else:
            raise ValueError("Invalid database type configured: %s" % augur.settings.main.datastores.main.type)

        if __is_bound:
            print("Database Configuration:")
            print("Type: %s" % dbtype)
            print("Target: %s" % dbtarget)

            db.generate_mapping(create_tables=True)
        else:
            print("No valid database configuration found")



import datetime
from munch import Munch, munchify
import logging
from augur import common
import pandas


def df_get_timing_status_keys(context, status_types):
    """
    Convert in progress statuses to keys that looks like this: "time_in_progress" (for example)
    :param status_types:
    :param context: The AugurContext object
    :return: An array of keys that look like this "time_in_progress" and always ends with "total_time_to_complete"
    """
    if isinstance(status_types,str):
        status_types = [status_types]

    statuses = []
    for tp in status_types:
        if tp == 'in progress':
            statuses.extend(context.workflow.in_progress_statuses())
        elif tp == 'done':
            statuses.extend(context.workflow.done_statuses())
        elif tp == 'open':
            statuses.extend(context.workflow.open_statuses())

    time_in_status_keys = ["_time_%s" % common.status_to_dict_key(x.tool_issue_status_name) for x in statuses]
    time_in_status_keys.append('_time_total_time_seconds')
    return time_in_status_keys


class Metrics(object):
    def __init__(self, context):
        self.context = context
        self.logger = logging.getLogger("augurjira")


class IssueCollectionMetrics(Metrics):

    def __init__(self, context, collection):
        self.collection = collection
        super(IssueCollectionMetrics, self).__init__(context)

    def status_analysis(self, options=None):
        status_counts = {
            'remaining_ticket_count': 0
        }

        for issue in self.collection:
            status_as_key = common.status_to_dict_key(issue.status)
            if status_as_key not in status_counts:
                status_counts[status_as_key] = 0
            status_counts[status_as_key] += 1

            if not self.context.workflow.is_resolved(issue.status, issue.resolution):
                status_counts['remaining_ticket_count'] += 1

        return munchify(status_counts)

    def timing_analysis(self, options=None):
        """
        For each issue in the collection, this will create an item a dictionary keyed on the issue key
        for each issue with the value being a dictionary containing the start, end and total times for each
        status along with the full in progress time.
        :param options: ----no options----
        :return: As follows:
            {
                'statuses': {
                    'status1': {
                        total: <timedelta>,
                        start: <datetime>,
                        end: <datetime>,
                    },
                    ...
                },
                'total_in_seconds': <float>,
                'total_as_time_delta': <timedelta>
            }
        """

        issues_with_timing = {}
        all_issue_status_timing = {}
        for issue in self.collection:
            # initialize all the keys for stats
            timing = {
                'statuses': {},
                'total_in_seconds': 0.0,
                'total_as_time_delta': None
            }
            for s in self.context.workflow.in_progress_statuses():
                s_as_key = common.status_to_dict_key(s)
                t = common.get_issue_status_timing_info(issue.issue, s)
                timing['statuses'][s_as_key] = {
                    'total': t['total_time'],
                    'start': t['start_time'],
                    'end': t['end_time']
                }
                timing['total_in_seconds'] += t['total_time'].total_seconds()

                if s_as_key not in all_issue_status_timing:
                    all_issue_status_timing[s_as_key] = 0

                all_issue_status_timing[s_as_key] += t['total_time'].total_seconds()

            timing['total_as_time_delta'] = datetime.timedelta(seconds=timing['total_in_seconds'])
            issues_with_timing[issue.key] = (munchify(timing))

        return munchify({
            'issues': issues_with_timing,
            'statuses': all_issue_status_timing
        })

    def point_analysis(self, options=None):
        """
        Does a very general analysis of a collection of issues
        :param options:
                - total_only (Boolean): If True, then issue details will not be included in many of the results -
                        just the totals
        :return: As follows:
        {
            'tickets': {
                'total_ticket_count': <integer>,
                'incomplete_ticket_count': <integer>,
                'unpointed_ticket_count': <integer>,
            }
            'points': {
                "total_points": 0.0,
                "completed_points": 0.0,
                "incomplete_points": 0.0,
                "percent_complete_points": 0,
                "abandoned_points": 0.0,
            },
            'developers': {
                'username': {
                    "info": <username>,
                    "complete_points": 0,
                    "incomplete_points": 0,
                    "abandoned_points": 0,
                    "percent_complete_points": 0,
                    'issues': []
                },
                ...
            }
        }
        """

        options = munchify(options)

        # Initialize the general analytics
        result = Munch({
            "ticket_count": self.collection.count(),
            "remaining_ticket_count": 0,
            "unpointed": 0.0,
            'developer_stats': {},
            'issues': {},
        })

        # Initialize the status counters
        result.update({common.status_to_dict_key(x): 0 for x in self.context.workflow.statuses})

        # Initialize point totals
        result.update({
            "complete": 0.0,
            "incomplete": 0.0,
            "total_points": 0.0,
            "percent_complete": 0,
            "abandoned": 0.0,
        })

        for issue in self.collection:
            assignee_cleaned = common.clean_username(issue.assignee)

            if issue.assignee not in result['developer_stats']:
                result['developer_stats'][assignee_cleaned] = Munch({
                    "info": issue.assignee,
                    "complete": 0,
                    "incomplete": 0,
                    "abandoned": 0,
                    "percent_complete": 0,
                    'issues': []
                })

            # Add this issue to the list of issues for the user
            result['developer_stats'][assignee_cleaned]['issues'].append(issue.key)

            if self.context.workflow.is_resolved(issue.status, issue.resolution):
                result['complete'] += issue.points
                result['developer_stats'][assignee_cleaned]['complete'] += issue.points

            elif self.context.workflow.is_abandoned(issue.status, issue.resolution):
                result["abandoned"] += issue.points
                result['developer_stats'][assignee_cleaned]['abandoned'] += issue.points

            else:
                result["incomplete"] += issue.points
                result['developer_stats'][assignee_cleaned]['incomplete'] += issue.points

            if not issue.points and not self.context.workflow.is_abandoned(issue.status, issue.resolution):
                result['unpointed'] += 1

        total_points = result['complete'] + result['incomplete'] + result['abandoned']
        result["percent_complete"] = int(((result['complete'] / total_points) if total_points > 0 else 0) * 100.0)
        result['total_points'] = total_points

        # get dev specific stats
        for assignee, dev in result['developer_stats'].items():
            total = dev['complete'] + dev['incomplete'] + dev['abandoned']
            dev['percent_complete'] = int((dev['complete'] / total if total > 0 else 0) * 100.0)
            dev['total_points'] = total

        return munchify(result)

    def get_data_frame(self, data_to_include=()):
        """
        Creates a data frame out of the issues in the collection.  You can choose what information to include in
        the frame.  By default, it will contain the following:
            * issuetype (string)
            * assignee  (string)
            * points (float)
            * dev_team (string)
            * description_length (int)
            * reporter (string)

        If you include 'timing' you also get a series of values that are prefixed with _timing_

        :param data_to_include: This is a list or tuple containing zero or more of the following keys:
            timing, [more to come]

        :return: Returns a pandas DataFrame
        """
        data = []

        timing_analysis = None

        if 'timing' in data_to_include:
            timing_analysis = self.timing_analysis()

        for issue in self.collection:
            row = {
                "key":issue.key,
                "issuetype": issue.issuetype,
                "assignee": issue.assignee,
                "points": issue.points,
                "dev_team": issue.team_name,
                "description_length": len(issue.description),
                "reporter": issue.reporter
            }

            if timing_analysis:

                if issue.key in timing_analysis.issues:
                    timing = {"_time_%s" % k: v['total'].total_seconds() for k, v in iter(timing_analysis.issues[issue.key].statuses.items())}
                    row.update(timing)
                    row["_time_total_time_seconds"] = timing_analysis.issues[issue.key].total_in_seconds

            data.append(row)

        return pandas.DataFrame(data=data)


class BoardMetrics(Metrics):
    def __init__(self, context, board):
        self._board = board
        super(BoardMetrics, self).__init__(context)

    def backlog_analysis(self, options=None):

        if not self._board or not self._board.option('include_sprint_reports'):
            self.logger.error("Board Metrics: The board object given must include sprint reports in order to "
                              "generate metrics")
            return False

        collection = self._board.get_backlog()

        metrics = {
            "points": {
                "unpointed": [],
                # pointed stories will be grouped by number of points keyed on their point value converted to string
            },
            "grade": ""
        }

        pointed = 0
        unpointed = 0

        for issue in collection.issues:

            if not issue.points:
                metrics['points']['unpointed'].append(issue.key)
                unpointed += 1
            else:
                pointed += 1
                key = str(issue.points)
                if key not in metrics['points']:
                    metrics['points'][key] = []

                metrics['points'][key].append(issue.key)

        total = pointed + unpointed
        if total > 0:
            percentage = (float(pointed) / float(total)) * 100.0
        else:
            percentage = 0.0

        metrics['total_unpointed_tickets'] = unpointed
        metrics['total_pointed_tickets'] = pointed
        metrics['pointed_percentage'] = percentage

        # now look at velocity for that board
        # sprints = self.get_sprints_from_boardby_team(team)

        if percentage >= 90.0:
            grade = "A"
        elif percentage >= 80.0:
            grade = "B"
        elif percentage >= 70.0:
            grade = "C"
        elif percentage >= 60.0:
            grade = "D"
        else:
            grade = "E"

        metrics['grade'] = grade

        return munchify(metrics)

    def historic_sprint_analysis(self, options=None):
        """
        Analyzes all the sprints in the given board.
        :param options: No options at this time.
        :return: Returns a munchified dictionary containing aggregate metrics.
        """

        sprint_collection = self._board.get_sprints()
        overall_metrics_list = []
        for sprint in sprint_collection:
            tp = (sprint.name, sprint.completed_points, sprint.incomplete_points, sprint.average_point_size)
            overall_metrics_list.append(tp)

        df_sprints = pandas.DataFrame(overall_metrics_list, columns=[
            "Name", "Points Completed", "Incomplete Points", "Average Point Size"])

        avg_velocity = df_sprints["Points Completed"].mean()
        low_velocity = df_sprints["Points Completed"].min()
        high_velocity = df_sprints["Points Completed"].max()
        avg_point_size = df_sprints["Average Point Size"].mean()
        highest_avg_point_size = df_sprints["Average Point Size"].max()
        lowest_avg_point_size = df_sprints["Average Point Size"].min()

        return munchify({
            "avg_velocity": avg_velocity,
            "low_velocity": low_velocity,
            "high_velocity": high_velocity,
            "avg_point_size": avg_point_size,
            "highest_avg_point_size": highest_avg_point_size,
            "lowest_avg_point_size": lowest_avg_point_size
        })



from augur.models.db import db as augur_db


def model_required(func):
    def inner(self, *args, **kwargs):
        if not self._model:
            raise ValueError("Model is required for this method to function properly")
        func(*args, **kwargs)

    return inner


def model_to_proxy(model):
    """Converts the given model to its proxy
    
    Arguments:
        model {db.Entity} -- The model to convert.  
        If not a model, the parameter will be returned as is.
    
    Returns:
        [*] -- Returns the proxy object if one was found, otherwise
                it returns the parameter given.
    """
    if not isinstance(model,augur_db.Entity):
        return model
    if hasattr(model,'_proxy_class'):
        proxy_name = model._proxy_class
        module = __import__('proxies')
        class_ = getattr(module, proxy_name)
        return class_(model)
    else:
        return model
                

def models_to_proxies (func):
    """Automatically converts models to proxies when a proxy is available
    
    Arguments:
        func {function} -- The wrapped function
    
    Returns:
        [function] -- The wrapper function
    """
    def inner(*args, **kwargs):
        for a in args:
            a = model_to_proxy(a)

        args = [model_to_proxy(a) for a in args]
        kwargs = {k:model_to_proxy(v) for k,v in kwargs}
            
        return func(*args,**kwargs)
    
    return inner
                
                            
class ProxyBase(object):
    
    def __init__(self, *args, **kwargs):
        self._model = None
        model_type = self._get_model_type()
        if 'model' in kwargs and model_type:
            if isinstance(kwargs['model'], model_type):
                self._model = kwargs['model']
            else:
                raise ValueError("The model given is not of the correct type.  We are expecting a model of type %s" % self._get_model().__name__)
        elif 'model_id' in kwargs and model_type:
            self._model = model_type[kwargs['model_id']]

    def _get_model_type(self):
        """Gets the class representing the model associated with this proxy
        
        Returns:
            [type] -- The model class or None
        """

        if not hasattr(self, 'model_type'):
            self.model_type = None
            if hasattr(self,'_model_class'):
                model_name = self._model_class
                module = __import__('models')
                self.model_type = getattr(module, model_name)
        return self.model_type                

    def pre_model_create(self):
        """Called just before a model is created
        """
        pass
                
    def post_model_create(self, model):
        """Called just after a model is created
        
        Arguments:
            model {db.Entity} -- The model object that was created
        """
        pass

    def create_model(self, *args, **kwargs):
        """Creates an instance of the model initializing it with the given params
        
        Raises:
            ValueError -- If this proxy does not define a model_class property a ValueException will be raised.
        """

        model_type = self._get_model_type()
        if model_type:
            self.pre_model_create()
            self._model = model_type(*args, **kwargs)
            if self._model:
                self.post_model_create(self._model)

            return self._model
        else:
            self.set_error("Unable to create a model because this proxy does not implement the _model_class property")
            return None

    @property
    def model(self):
        return self._model

    @property
    def last_error(self):
        return self.last_error

    def set_error(self, error_string):
        self.last_error = error_string  



from marshmallow import Schema, fields, ValidationError, pprint, post_load
import augur


def validate_resolution_types(s):
    if s not in augur.db.TOOL_ISSUE_RESOLUTION_TYPES:
        raise ValidationError("{} is not a valid resolution type".format(s))


def validate_issue_status_type(s):
    if s not in augur.db.TOOL_ISSUE_STATUS_TYPES:
        raise ValidationError("{} is not a valid issue status type".format(s))


def validate_issue_type_types(s):
    if s not in augur.db.TOOL_ISSUE_TYPE_TYPES:
        raise ValidationError("{} is not a valid issue type type".format(s))


def validate_staff_roles(s):
    if s not in augur.db.ROLES:
        raise ValidationError("{} is not a valid staff role".format(s))


def validate_staff_types(s):
    if s not in augur.db.STAFF_TYPES:
        raise ValidationError("{} is not a valid staff type".format(s))


def validate_staff_status(s):
    if s not in augur.db.STATUS:
        raise ValidationError("{} is not a valid staff status".format(s))


def validate_notification_targets(s):
    if s:
        targets = s.split(",")
    else:
        return

    for t in targets:
        if t.trim() not in augur.db.NOTIFY_TYPES:
            raise ValidationError("{} is not a valid notification target".format(t))


def validate_build_types(s):
    if s:
        types = s.split(",")
    else:
        return

    for t in types:
        if t.trim() not in augur.db.BUILD_TYPES:
            raise ValidationError("{} is not a valid build type".format(t))


class ToolIssueResolutionSchema (Schema):
    id = fields.Integer()
    tool_issue_resolution_name = fields.String()
    tool_issue_resolution_type = fields.String(validate=validate_resolution_types)


class ToolIssueStatusSchema (Schema):
    id = fields.Integer()
    tool_issue_status_name = fields.String()
    tool_issue_status_type = fields.String(validate=validate_issue_status_type)


class ToolIssueTypeSchema (Schema):
    id = fields.Integer()
    tool_issue_type_name = fields.String()
    tool_issue_type_type = fields.String(validate=validate_issue_type_types)


class VendorSchema (Schema):
    id = fields.Integer()
    name = fields.String()
    engagement_contact_first_name = fields.String()
    engagement_contact_last_name = fields.String()
    engagement_contact_email = fields.Email()
    billing_contact_first_name = fields.String()
    billing_contact_last_name = fields.String()
    billing_contact_email = fields.Email()
    tempo_id = fields.Integer()
    consultants = fields.Nested('StaffSchema', many=True)


class ProductSchema (Schema):
    id = fields.Integer()
    name = fields.String()
    key = fields.String()
    teams = fields.Nested('TeamSchema', many=True)


class ToolProjectSchema (Schema):
    id = fields.Integer()
    tool_project_key = fields.String()


class ToolProjectCategorySchema (Schema):
    id = fields.Integer()
    tool_category_name = fields.String()


class StaffSchema (Schema):
    id = fields.Integer()
    first_name = fields.String()
    last_name = fields.String()
    company = fields.String()
    avatar_url = fields.Url()
    role = fields.String(validate=validate_staff_roles)
    email = fields.Email()
    rate = fields.Float()
    start_date = fields.DateTime()
    type = fields.String(validate=validate_staff_types)
    jira_username = fields.String()
    slack_id = fields.String()
    github_username = fields.String()
    status = fields.String(validate=validate_staff_status)
    teams = fields.Nested('TeamSchema', many=True, only=('id','name'))
    base_daily_cost = fields.Float()
    base_weekly_cost = fields.Float()
    base_annual_cost = fields.Float()
    vendor = fields.Nested('VendorSchema', only=('id', 'name'))
    notification = fields.Nested('NotificationSchema')

    @post_load
    def make_staff_entity(self, data):
        from augur.db import Staff
        if 'id' in data:
            # load an existing object and update with the given data.
            s = Staff[data['id']]
            data.pop('id')

            # If there is no data to update then we can skip the update.
            if data:
                s.set(**data)

            return s
        else:
            # create a new staff object
            return Staff(**data)


class AgileBoardSchema (Schema):
    id = fields.Integer()
    jira_id = fields.Integer()


class TeamSchema (Schema):
    id = fields.Integer()
    name = fields.String()
    members = fields.Nested('StaffSchema', only=('id', 'email', 'first_name', 'last_name', 'jira_username', 'github_username'), many=True)
    agile_board = fields.Nested('AgileBoardSchema')
    product = fields.Nested('ProductSchema', only=('id', 'name', 'key'))
    groups = fields.Nested('GroupSchema', only=('id', 'name', 'workflow.name', 'workflow.id'), many=True)
    notification = fields.Nested('NotificationSchema')


class NotificationSchema (Schema):
    id = fields.Integer()
    build = fields.String(validate=validate_notification_targets)
    deploy = fields.String(validate=validate_notification_targets)
    build_types = fields.String(validate=validate_build_types)


class WorkflowDefectProjectFilterSchema (Schema):
    id = fields.Integer()
    project_key = fields.String()
    issue_types = fields.Nested('ToolIssueTypeSchema', many=True)


class GroupSchema (Schema):

    id = fields.Integer()
    name = fields.String()
    workflow = fields.Nested('WorkflowSchema', only=['id','name','statuses','resolutions','projects','categories',
                                                     'issue_types','defect_projects'], many=False)
    products = fields.Nested(ProductSchema, only=['id','name', 'key', 'teams.id','teams.name'], many=True)
    teams = fields.Nested(TeamSchema, only=['id', 'name', 'members.id', 'members.first_name',
                                            'members.last_name', 'members.email'], many=True)


class WorkflowSchema (Schema):
    id = fields.Integer()
    name = fields.String()
    statuses = fields.Nested(ToolIssueStatusSchema, many=True)
    resolutions = fields.Nested(ToolIssueResolutionSchema, many=True)
    projects = fields.Nested(ToolProjectSchema, many=True)
    categories = fields.Nested(ToolProjectCategorySchema, many=True)
    issue_types = fields.Nested(ToolIssueTypeSchema, many=True)
    defect_projects = fields.Nested(WorkflowDefectProjectFilterSchema, many=True)
    groups = fields.Nested(GroupSchema, many=True)





class AugurContext(object):
    """
    This contains information that is used by the Augur library to identify constraints that should be
    used when requesting data.  The Context object is defined by a "Group".  Groups
    are associated with a workflow, teams and other information.  Many functions within
    augur will require a context object in order to know how to filter and interpret data.
    """

    def __init__(self, group_id):
        from augur import get_group
        self._group = get_group(group_id)
        self._workflow = self.group.workflow

    @property
    def workflow(self):
        return self._workflow

    @property
    def group(self):
        return self._group


import json
import os
import re

import arrow
import datetime
import pytz
from munch import munchify

import augur
from augur import db
from functools import reduce

__author__ = 'karim'

from math import sqrt, floor
from dateutil.parser import parse

JIRA_KEY_REGEX = r"([A-Za-z]+\-\d{1,6})"

# Note: The order matters in this list.  Time based matches are first to ensure that
#   the time is not truncated in cases where date matches are found then the rest of the
#   list is skipped.
POSSIBLE_DATE_TIME_FORMATS = [
    "YYYY-MM-DD HH:mm",
    "MM-DD-YYYY HH:mm",
    "MM/DD/YYYY HH:mm",
    "YYYY/MM/DD HH:mm",
    "MM/DD/YYYY HH:mm",
    "M/D/YYYY HH:mm",
    "M/D/YY HH:mm",
]

SITE_ROOT = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))


class Struct:
    """
        Create objects out of dictionaries like this:
        >>> d = {"test":1,"foo":"bar"}
        >>> s = Struct(**d)
        >>> s.test
    """
    def __init__(self, **entries):
        self.__dict__.update(entries)


class AugurJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        from augur.context import AugurContext

        if isinstance(obj, datetime.timedelta):
            return obj.total_seconds()
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, Resource):
            return obj.raw
        elif isinstance(obj, AugurContext):
            return {
                "group_id": obj.group.id
            }

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def remove_null_fields(d):
    """
    Takes a dictionary and returns a new dictionary with fields that have null values.
    :param d: A dictionary to remove null fields from
    :return:
    """
    return dict((k, v) for k, v in iter(d.items()) if v is not None)


def clean_issue(issue):
    """
    This will put commonly used fields at the root of the dictionary and remove all of the custom fields that are
    empty.
    :param issue: A dict that is the the issue to clean
    :return: Returns a dictionary.
    """
    points_field_name = augur.api.get_issue_field_from_custom_name('Story Points')
    points = augur.common.deep_get(issue, 'fields', points_field_name)
    status = augur.common.deep_get(issue, 'fields', 'status', 'name') or ''
    resolution = augur.common.deep_get(issue, 'fields', 'resolution', 'name') or ''
    return {
        'key': issue['key'],
        'summary': augur.common.deep_get(issue, 'fields', 'summary'),
        'assignee': augur.common.deep_get(issue, 'fields', 'assignee', 'name') or 'unassigned',
        'description': augur.common.deep_get(issue, 'fields', 'description'),
        'fields': remove_null_fields(issue['fields']),
        'points': float(points if points else 0.0),
        'status': status.lower(),
        'changelog': augur.common.deep_get(issue,'changelog'),
        'resolution': resolution.lower(),
    }


def find_team_name_in_string(team_name, string_to_search):
    """
    Utility that searches the given string for the given team name.  It does the work of stripping out the
    word "Team" from the team_name for you so it's more permissive than just doing a substring search.

    :param team_name: The name of the team
    :param string_to_search: The source string to search for the team name
    :return: Returns the True if the name was found, False otherwise.
    """
    # Remove the word "Team" from the team name (if necessary)
    team_replace = re.compile(re.escape('team'), re.IGNORECASE)
    team_name = team_replace.sub('', team_name).strip()
    return team_name.lower() in string_to_search.lower()


def utc_to_local(utc_dt):
    local_tz = pytz.timezone('America/New_York')
    return utc_dt.replace(tzinfo=pytz.utc).astimezone(local_tz)


def extract_jira_tickets(text):
    """
    Gets a list of all the JIRA keys found in the given text
    :param text: The string to search
    :return: A list of strings
    """
    match = re.search(JIRA_KEY_REGEX,text)
    return match.groups() or []


def get_week_range(date):
    """
    This will return the start and end of the week in which the given date resides.
    :param date: The source date
    :return: A tuple containing the start and end date of the week (in that order)
    """
    start = date - datetime.timedelta(days=date.weekday())
    end = start + datetime.timedelta(days=6)
    return start,end


def standard_deviation(lst,population=True):

    sd = 0

    try:
        num_items = len(lst)

        if num_items == 0:
            return 0

        mean = sum(lst) / num_items
        differences = [x - mean for x in lst]
        sq_differences = [d ** 2 for d in differences]
        ssd = sum(sq_differences)

        if population is True:
            variance = ssd / num_items
        elif num_items > 1:
            variance = ssd / (num_items - 1)
        else:
            variance = 0

        sd = sqrt(variance)

    except Exception as e:
        print("Problem during calculation of standard deviation: %s"%e.message)

    return sd


def get_date_range_from_strings(start,end, default_start=None, default_end=None):
    if not start:
        # Don't go back further than 90 days for the sake of performance and storage.
        start = default_start
    else:
        start = arrow.get(start, "YYYY-MM-DD").floor('day')

    if not end:
        end = default_end
    else:
        end = arrow.get(end, "YYYY-MM-DD").ceil('day')

    return start,end


def deep_get(dictionary, *keys):
    """
    Retrieves a deeply nested dictionary key checking for existing keys along the way.  
    Returns None if any key is not found.
    :param dictionary: The dictionary to retrieve the data from.
    :param default: The default value to return if there is a problem during the search
    :param keys: Ordered parameters representing the keys (in the order they should be referenced)
    :return: Returns the value if found, None otherwise.
    """
    default = None
    return reduce(lambda d, key: d.get(key) if d else default, keys, dictionary)


def status_to_dict_key(status):
    if isinstance(status, db.ToolIssueStatus):
        status_str = status.tool_issue_status_name
    else:
        status_str = status

    return "%s" % status_str.lower().replace(" ", "_")


def clean_username(username):
    """
    This cleans up a username to ensure it doesn't have any characters that would interfere with things
    like mongo storage, dict key use, database storage, etc.
    :param username:
    :return:
    """
    return username.replace(".", "_")


def get_issue_status_timing_info(issue, status):
    """
    Gets a single tickets timing information including when in started, ended and the total time in the status.
    :param issue: The ticket in dictionary form
    :param status: The ToolIssueStatus to look for.
    :return: Returns a dict containing:
                start_time: datetime when the issue first started in the status
                end_time: datetime when the issue last left the status
                total_time: timedelta with the total time in status
    """
    status_name = status.tool_issue_status_name
    track_time = None
    total_time = datetime.timedelta()

    if 'changelog' not in issue or not issue['changelog']:
        return {
            "start_time":None,
            "end_time":None,
            "total_time":datetime.timedelta(seconds=0)
        }

    history_list = issue['changelog']['histories']

    # added sorting past > present because the API is inconsistently ordering the results.
    history_list.sort(key=lambda x: x['id'], reverse=False)

    start_time = None
    for history in history_list:
        items = history['items']

        for item in items:
            if item['field'] == 'status' and item['toString'].lower() == status_name.lower():
                # start status
                track_time = parse(history['created'])
                if not start_time:
                    start_time = track_time

                break
            elif track_time and item['field'] == 'status' and item['fromString'].lower() == status_name.lower():
                # end status
                if track_time:
                    # only recalculate if track_time has a value.  It can happen that track_time has no
                    # value if the API only returns X number of historical items and the
                    total_time += (parse(history['created']) - track_time)
                track_time = None
                break

    if track_time and not total_time:
        # In this case the issue is currently in the requested status which means we need to set the "end" time to
        #   NOW because there's no record of the *next* status to subtract from.
        end_time = utc_to_local(datetime.datetime.now())
        total_time = end_time - track_time
    else:
        end_time = track_time

    return munchify({
        "start_time":start_time,
        "end_time":end_time,
        "total_time":total_time
    })


def get_date_from_week_number(week_number):
    year = datetime.datetime.now().year
    conversion_str = "%d-W%d" % (year, int(week_number))
    return datetime.datetime.strptime(conversion_str + '-1', "%Y-W%W-%w")


def format_timedelta(value, time_format="{days} days, {hours2}:{minutes2}:{seconds2}", time_format_no_days="{hours}h"):
    """
    Formats timedelta and uses the following options for the formatting string:
        seconds:        Seconds, no padding
        seconds2:       Seconds with zero padding
        minutes:        Minutes, no padding
        minutes:        Minutes with zero padding
        hours:          Hours, no padding
        hours2:         Hours with zero padding
        days:           Days (when used in conjuction with other units - not total)
        years:          Years (when used in conjuction with other units - not total)
        seconds_total:  The total number of seconds
        minutes_total   The total number of minutes
        hours_total:    The total number of hours
        days_total:     The total number of days
        years_total:    The total number of years

        Example:
            "{days} days, {hours2}:{minutes2}:{seconds2}"
        will format into something like:
            "3 days, 02:20:00"
    Args:
        value(datetime.timedelta): The value to format
        time_format(str): The patter to format into (see description for options)
    """
    if hasattr(value, 'seconds'):
        seconds = value.seconds + value.days * 24 * 3600
    else:
        seconds = int(value)

    seconds_total = seconds

    minutes = int(floor(seconds / 60))
    minutes_total = minutes
    seconds -= minutes * 60

    hours = int(floor(minutes / 60))
    hours_total = hours
    minutes -= hours * 60

    days = int(floor(hours / 24))
    days_total = days
    hours -= days * 24

    years = int(floor(days / 365))
    years_total = years
    days -= years * 365

    if days_total > 0:
        return time_format.format(**{
            'seconds': seconds,
            'seconds2': str(seconds).zfill(2),
            'minutes': minutes,
            'minutes2': str(minutes).zfill(2),
            'hours': hours,
            'hours2': str(hours).zfill(2),
            'days': days,
            'years': years,
            'seconds_total': seconds_total,
            'minutes_total': minutes_total,
            'hours_total': hours_total,
            'days_total': days_total,
            'years_total': years_total,
        })
    else:
        return time_format_no_days.format(**{
            'seconds': seconds,
            'seconds2': str(seconds).zfill(2),
            'minutes': minutes,
            'minutes2': str(minutes).zfill(2),
            'hours': hours,
            'hours2': str(hours).zfill(2),
            'days': days,
            'years': years,
            'seconds_total': seconds_total,
            'minutes_total': minutes_total,
            'hours_total': hours_total,
            'days_total': days_total,
            'years_total': years_total
        })



import datetime
import exceptions
import json
import logging
import re

import dateutil
import github
import pytz
import yaml
from github import Github
from github.GithubObject import GithubObject

from augur import settings
from augur.api import get_jira
import augur.api

DEFAULT_LOOKBACK_DAYS = 90


class GitFileNotFoundError(exceptions.Exception):
    """Used for reporting a problem when scanning a repo for a file"""
    pass


class AugurGithubDevStats(object):
    def __init__(self, username):
        self.user = username
        self.prs = []
        self.avg_changed_files_per_pr = 0
        self.avg_comments_per_pr = 0
        self.highest_changes_in_pr = 0
        self.highest_comments_in_pr = 0
        self.avg_length_of_time_pr_was_open = datetime.timedelta()
        self.dirty = False

    def reset(self):
        self.__init__(self.user)

    def add_pr(self, pr):

        if pr['user']['login'] == self.user:
            self.dirty = True

            # we need a version of this attribute that doesn't have the underscore so that we can
            #   display its values in a django template.
            pr['links'] = pr['pull_request']
            self.prs.append(pr)

            comment_count = pr['comments']
            self.avg_comments_per_pr += comment_count
            if comment_count > self.highest_comments_in_pr:
                self.highest_comments_in_pr = comment_count

            closed_at = pr['closed_at']
            created_at = pr['created_at']

            if not isinstance(pr['closed_at'], datetime.datetime):
                closed_at = dateutil.parser.parse(pr['closed_at'])
            closed_at = closed_at.replace(tzinfo=None)

            if not isinstance(pr['created_at'], datetime.datetime):
                created_at = dateutil.parser.parse(
                    pr['created_at']).replace(tzinfo=pytz.UTC)
            created_at = created_at.replace(tzinfo=None)

            if pr['state'] in ['merged', 'closed']:
                self.avg_length_of_time_pr_was_open += (closed_at - created_at)

    def as_dict(self):

        if self.dirty:
            self.avg_changed_files_per_pr /= len(self.prs)
            self.avg_comments_per_pr /= len(self.prs)

            avg_secs = self.avg_length_of_time_pr_was_open.total_seconds() / len(self.prs)
            self.avg_length_of_time_pr_was_open = datetime.timedelta(
                seconds=avg_secs)
            self.dirty = False

        return {
            'prs': self.prs,
            'avg_changed_files_per_pr': self.avg_changed_files_per_pr,
            'avg_comments_per_pr': self.avg_comments_per_pr,
            'highest_changes_in_pr': self.highest_changes_in_pr,
            'highest_comments_in_pr': self.highest_comments_in_pr,
            'avg_length_of_time_pr_was_open': self.avg_length_of_time_pr_was_open,
        }


class AugurGithub(object):
    def __init__(self):
        self.github = Github(login_or_token=settings.main.integrations.github.login_token,
                             password="x-oauth-basic",
                             base_url=settings.main.integrations.github.base_url,
                             per_page=200)

        self.jira = get_jira()
        self.logger = logging.getLogger("augurgithub")

    def fetch_further_reviews(self, org):
        """
        Gets all the further review information for each repo in the given org
        :param org: The name of the org to return further review info from
        :return: Returns a dictionary containing further review information
        """
        repos = self.get_repos_in_org(org)
        org_further_reviews = {}
        for r in repos:
            org_further_reviews[r.name] = self.get_repo_further_review(r)

        return org_further_reviews

    def get_commit_info(self, commit_hash, org_repo=None):
        """
        Gets details about a single commit based on the given commit hash.  If there is more than one returned
        from github, the first is returned from this function
        :param commit_hash: The sha1 hash of the commit (shortened or full length)
        :param org_repo: The name of the org/repo combo (<org>/<repo>) to search. When you specify the repo, it
                            gives you better results.
        :return: Return github.Commit.Commit
        """

        if not org_repo:
            commits = self.github.search_commits(query="hash:%s" % commit_hash)
            if commits.totalCount > 0:
                return commits[0]
            else:
                logging.warning(
                    "Unable to find the following commit: %s" % commit_hash)
                return None
        else:
            try:
                org, repo = org_repo.split("/")
            except ValueError:
                logging.error(
                    "Got an invalid org_repo parameter in get_commit_info")
                return None

            org_ob, repo_ob = self.get_org_and_repo_from_params(repo, org)
            if repo_ob:
                commit = repo_ob.get_commit(commit_hash)
                if not commit:
                    logging.warning(
                        "Unable to find the following commit (with repo): %s" % commit_hash)
                    return None
                else:
                    return commit
            else:
                logging.error(
                    "Unable to find the given repo (%s) or org (%s)" % (repo, org))
                return None

    def _prepare_pr_search(self, orgs, state, since, sort, order):
        """
        Prepare the basic query for a PR search including state, beginning date, orgs/repos
        It returns a string containing the full query and caller can add to it to filter further
        :param order: The sort order can be asc or desc
        :param sort: The sort order.  Can be comments, created, or updated
        :param state: The state of the pr (open, closed, merged)
        :param org: The name of the organization (as a string)
        :param since: The earliest date when the PRs were first opened
        :return: Returns a query string (just the q parameter's value
        """
        # Convert to list if it's not already
        query = " type:pr"

        if orgs:
            orgs = orgs if isinstance(orgs, list) else [orgs]

            local_repo_obs = []
            for o in orgs:
                cached_repos = augur.api.get_memory_cached_data(
                    "repos_for_%s" % o)
                if not cached_repos:
                    repos = [r.raw_data for r in self.get_repos_in_org(o)]
                    augur.api.memory_cache_data(
                        {'data': repos}, "repos_for_%s" % o)
                else:
                    repos = cached_repos[0]['data']

                local_repo_obs += repos

            # build a list of all the repos to search (no way to filter by org apparently)
            query += " " + \
                     " ".join(["repo:%s/%s" % (r['organization']['login'], r['name'])
                               for r in local_repo_obs])

        if state in ("open", "closed"):
            query += " is:%s" % state

        if since:
            if state == 'merged':
                query += ' merged:>%s' % since.date().isoformat()
            else:
                query += ' created:>%s' % since.date().isoformat()

        return query

    def fetch_user_data(self, user, lookback_days=60):
        """
        Get github status for a single user.  This will retrieve the most recent data if not already
        stored in the database.  So at this point, this call could take 20 minutes to complete
        :param user: The name of the user
        :param lookback_days: The number of days to look back in time for PRs and other data
        :return: Returns a dict with github stats info in it
        """
        dev_stats_ob = AugurGithubDevStats(user)
        since = datetime.datetime.now() - datetime.timedelta(days=int(lookback_days))
        merged_prs = self.fetch_author_merged_prs(user, since)
        for pr in merged_prs:
            dev_stats_ob.add_pr(pr)

        return dev_stats_ob.as_dict()

    def fetch_author_merged_prs(self, username, since, sort="created", order="desc"):
        """
        This will retrieve one or more authors prs in the given state
        :param username: A string or list of usernames
        :param state: Can be one of open, closed, merged
        :param sort: The sort order can be one of comments, created or updated
        :param order: Can be one of asc or desc
        :return: Returns a list of search results as a list of dictionaries
        """
        query = self._prepare_pr_search(
            orgs=None, state="merged", since=since, sort=sort, order=order)
        query += " author:%s" % username
        results = self.github.search_issues(
            query=query, sort=sort, order=order)
        return [r.raw_data for r in results]

    def fetch_organization_members(self, organization):
        """
        Gets a list of all organization members as a lits of dictionaries
        :param organization: The name of the organization
        :return: Returns a list of organization Member objects
        """
        org_ob = self.get_organization(organization)
        return [m.raw_data for m in org_ob.get_members()] if org_ob else []

    def get_pr(self, org, repo, number):
        """
        Gets the PR object from the given org/repo and PR number.
        :param org: The org string or Organization object
        :param repo: The repo string or Repo object
        :param number: The PR number
        :return: The PullRequest object.
        """
        try:
            org_ob, repo_ob = self.get_org_and_repo_from_params(repo, org)
            return repo_ob.get_pull(number)
        except TypeError:
            logging.error("Could not get org and repo objects from given data")
            return None

    def get_organization(self, org):
        """
        Takes an organization name and returns an organization object
        :param org: The organization's name
        :return: The Organization object or None
        """
        try:
            # first we try getting a real organization
            org_ob = self.github.get_organization(org)
        except github.UnknownObjectException:
            # try a user org if there is not actual org
            org_ob = self.github.get_user(org)

        return org_ob

    def get_repos_in_org(self, org):
        """
        Gets a list of all the repositories in the form of Repository objects
        from the given organization name.
        :param org: The name of an organization
        :return: Returns a list of Repository objects
        """
        org_ob = self.get_organization(org)
        return [r for r in org_ob.get_repos()] if org_ob else []

    def get_repo_commit_stats(self, repo, org=None):
        """
        Gets commits information for the last 90 days.  This includes things like:
            * A flat list of commits
            * the commit object and counts organized by author
            * total commits
            * total comments
            * average comments per commit
        :param repo: The repo object or name
        :param org: The org object or name.  Note that the org must be given if the repo is just a string
        :return: Returns a dictionary containing commit information
        """
        org_ob, repo_ob = self.get_org_and_repo_from_params(repo, org)

        since = datetime.datetime.now() - datetime.timedelta(days=90)
        commits = repo_ob.get_commits(since=since)
        by_author = dict()
        avg_comments = 0
        total_comments = 0

        commit_objects = []
        try:
            for c in commits:
                try:
                    if c.author.login not in by_author:
                        by_author[c.author.login] = {
                            "count": 0,
                            "commits": []
                        }
                    by_author[c.author.login]['count'] += 1
                    by_author[c.author.login]['commits'].append(c.raw_data)

                    comments = c.get_comments()
                    comment_objects = []
                    for cmt in comments:
                        comment_objects.append(cmt.raw_data)

                    total_comments += len(comment_objects)
                    commit_objects.append(c.raw_data)
                except Exception as e:
                    self.logger.error(
                        "Encountered error when reviewing commits for repo %s: %s" % (repo_ob.name, e.message))
                    continue
        except github.GithubException as e:
            self.logger.error(
                "Encountered an error while iterating over commits: %s" % e.message)

        total_commits = len(commit_objects)

        if total_commits:
            avg_comments = float(total_comments) / float(total_commits)

        return {
            "commits": commit_objects,
            "by_author": by_author,
            "total_commits": total_commits,
            "total_comments": total_comments,
            "avg_comments_per_commit": avg_comments
        }

    def get_org_and_repo_from_params(self, repo, org=None):
        """
        Tries to get the org and repo objects based on the given params.  If the repo is an object
        then it will just return that and the org object contained within.  If the repo param
        is a string then the org param must be non-null so that the specific repo can be found.
        In that case, the org can either be an Organization object or a string.
        :param repo: The repo object or name
        :param org: The org object or name.  Note that the org must be given if the repo is just a string
        :return: Returns two results: Repository object, Organization object.
        """
        if not org and not isinstance(repo, GithubObject):
            raise TypeError(
                "If repo given is not a github object then organization must be specified")

        repo_ob = None
        org_ob = None

        if repo and isinstance(repo, GithubObject):
            repo_ob = repo

        if org and isinstance(repo, GithubObject):
            raise TypeError(
                "If org is given then repo should be a string otherwise there's no reason to pass the org")

        if org and isinstance(org, str):
            org_ob = self.github.get_organization(org)

        elif org and isinstance(org, GithubObject):
            org_ob = org

        elif org:
            raise TypeError("Org must be a github object or a string")

        if not repo_ob:
            repo_ob = org_ob.get_repo(repo)

        if not org_ob and repo_ob:
            org_ob = repo_ob.organization

        return org_ob, repo_ob

    def get_org_component_data(self, org):
        """
        Gets all repos in an org and retrieves information about each that comes from github along with additional
        info including further review data, additional aggregate data and the readme summary.  Note that this
        will use cached data if available.

        :param org: The organization object or name
        :return: Returns a dict containing the data.
        """
        repos = self.get_repos_in_org(org) or []
        data = {
            'org': None,
            'repos': []
        }

        for repo in repos:

            further_review = self.get_repo_further_review(repo)
            readme_summary_list = self.get_repo_readme_summary(repo)

            repo.raw_data["further_review"] = further_review or {}
            repo.raw_data['readme_summary'] = readme_summary_list[0] if len(
                readme_summary_list) else ""
            repo.raw_data["ua_stats"] = self.get_repo_commit_stats(repo)

            data['repos'].append(repo.raw_data)
            if not data['org']:
                data['org'] = repo.owner.login

        return data

    def get_repo_readme_summary(self, repo, org=None):
        """
        Finds the text between the h1 and h2 in the readme file for a repo and returns it in the form of an array
        of strings (one for each line)
        :param repo: The repo object or name
        :param org: The org object or name.  Note that the org must be given if the repo is just a string
        :return: Returns a list of strings.
        """
        org_ob, repo_ob = self.get_org_and_repo_from_params(repo, org)
        result = []
        try:
            content_ob = repo_ob.get_file_contents("README.md")
            if content_ob:
                readme_str = content_ob.decoded_content
                if readme_str:
                    try:
                        result = re.findall(
                            '#.*\n([^#]*)', readme_str, re.MULTILINE)
                        return result
                    except Exception as e:
                        self.logger.info(
                            "Parse error during processing of README.md: %s" % e.message)
                else:
                    self.logger.warning("Unable to retrieve README.md")
            else:
                self.logger.warning("Unable to find README.md file")
        except github.UnknownObjectException as e:
            self.logger.error("Error occurred during README analysis: %s (%s)" % (
                e.message, str(e.__class__)))
        except github.GithubException as e:
            self.logger.error("Error occurred during further review analysis: %s (%s)" % (
                e.message, str(e.__class__)))
        return result

    def get_repo_further_review(self, repo, org=None):
        """
        Gets the further review information from the given repo
        :param repo: The repo object or name
        :param org: The org object or name.  Note that the org must be given if the repo is just a string
        :return: Returns a dictionary containing information about the maintainers and owners
        """
        fr = augur.api.get_memory_cached_data('FR_OPR_' + repo.full_name)
        if fr:
            return fr

        def parse_user(user_str):
            name_match = re.match(r"^([^\(<]+)", user_str)
            email_match = re.match(r".*<(.*)>.*", user_str)
            user_match = re.match(r".*\(@(.*)\).*", user_str)

            return {
                'name': name_match.group(1) if name_match else "",
                'email': email_match.group(1) if email_match else "",
                'username': user_match.group(1) if user_match else "",
            }

        org_ob, repo_ob = self.get_org_and_repo_from_params(repo, org)

        result = {
            "owner": None,
            "maintainers": [],
        }

        try:
            content_ob = repo_ob.get_file_contents(".further-review.yaml")
            if content_ob:
                yaml_str = content_ob.decoded_content
                if yaml_str:
                    try:
                        further_review = yaml.load(yaml_str)
                        if 'reviews' not in further_review or not isinstance(further_review['reviews'], list):
                            self.logger.warning(
                                "Unable to find reviews section in further-review file")
                        else:
                            for index, review in enumerate(further_review['reviews']):
                                if review['name'].lower() == 'general maintainers':
                                    reviewers = review['logins'] if review['logins'] else [
                                    ]
                                    for r in reviewers:
                                        user_ob = parse_user(r)
                                        result['maintainers'].append(user_ob)
                            if 'owner' in further_review and further_review['owner']:
                                result['owner'] = parse_user(
                                    further_review['owner'])

                    except yaml.YAMLError as e:
                        self.logger.info(
                            "YAML error during processing of .further-review.yaml: %s" % e.message)
                else:
                    self.logger.warning(
                        "Unable to retrieve .further-review.yaml")
            else:
                self.logger.warning("Unable to find further-review.yaml file")
        except github.UnknownObjectException as e:
            self.logger.error(
                "Got a github error indicating that we were unable to find a file in the repo: %s (%s)" % (
                    e.message, str(e.__class__)))
        except github.GithubException as e:
            self.logger.error("Error occurred during further review analysis: %s (%s)" % (
                e.message, str(e.__class__)))

        augur.api.memory_cache_data(result, 'FR_OPR_' + repo.full_name)

        return result

    def get_repo_package_json(self, repo, org=None):
        """
        Gets the package.json JSON from a given repo
        :param repo: The repo object or name
        :param org: The org object or name.  Note that the org must be given if the repo is just a string
        :return: Returns a dictionary object containing all of the package.json contents
        """
        org_ob, repo_ob = self.get_org_and_repo_from_params(repo, org)

        try:
            content_ob = repo_ob.get_file_contents("package.json")
            if content_ob:
                json_str = content_ob.decoded_content
                if json_str:
                    try:
                        package_ob = json.loads(json_str)
                        if 'maintainers' in package_ob and isinstance(package_ob['maintainers'], list):
                            for index, maintain in enumerate(package_ob['maintainers']):
                                if isinstance(maintain, str):
                                    parts = re.match(
                                        r"(?:(.*))\s(?:<?(.+@[^>]+)>?)?", maintain)
                                    groups = parts.groups()
                                    package_ob['maintainers'][index] = {
                                        'name': groups[0] if len(groups) > 0 else "",
                                        'email': groups[1] if len(groups) > 1 else ""
                                    }
                        if 'owner' in package_ob:
                            owner = package_ob['owner']
                            if isinstance(owner, str):
                                parts = re.match(
                                    r"(?:(.*))\s(?:<?(.+@[^>]+)>?)?", owner)
                                groups = parts.groups()
                                package_ob['owner'] = {
                                    'name': groups[0] if len(groups) > 0 else "",
                                    'email': groups[1] if len(groups) > 1 else ""
                                }
                        return package_ob

                    except ValueError as e:
                        raise GitFileNotFoundError(
                            "Package found but had invalid JSON in it %s: %s" % (repo.name, e.message))
                else:
                    raise GitFileNotFoundError(
                        "Found package file but nothing in it for repo %s" % repo.name)
            else:
                raise GitFileNotFoundError(
                    "Cannot find package file in %s" % repo.name)
        except github.UnknownObjectException as e:
            raise GitFileNotFoundError(
                "Cannot find package file in %s: %s" % (repo.name, e.message))



import logging

from jira import JIRA, Issue
from jira.resources import Board
from munch import munchify

from augur import api
from augur import settings


class AugurJira(object):
    """
    A thin wrapper around the Jira module providing some refinement for things like fields, convenience methods
    for ticket actions and awareness for Augur-specific data types.
    """
    jira = None

    def __init__(self, server=None, username=None, password=None):

        self.logger = logging.getLogger("augurjira")
        self.server = server or settings.main.integrations.jira.instance
        self.username = username or settings.main.integrations.jira.username
        self.password = password or settings.main.integrations.jira.password
        self.fields = None

        self.jira = JIRA(basic_auth=(
            self.username,
            self.password),
            server=self.server,
            options={"agile_rest_path": "agile"})

        self._field_map = {}

        self._default_fields = munchify({
            "summary": None,
            "description": None,
            "status": None,
            "priority": None,
            "parent": None,
            "resolution": None,
            "epic link": None,
            "dev team": None,
            "labels": None,
            "issuelinks": None,
            "development": None,
            "reporter": None,
            "assignee": None,
            "issuetype": None,
            "project": None,
            "creator": None,
            "attachment": None,
            "worklog": None,
            "story points": None,
            "changelog": None
        })

        self.fields = api.get_memory_cached_data('custom_fields')
        if not self.fields:
            fields = api.memory_cache_data(self.jira.fields(), 'custom_fields')
            self.fields = {f['name'].lower(): munchify(f) for f in fields}

        default_fields = {}
        for df, val in self._default_fields.items():
            default_fields[df] = self.get_field_by_name(df)

        self._default_fields = munchify(default_fields)

    @property
    def default_fields(self):
        """
        Returns a dict containing the friendly name of fields as keys and jira's proper field names as values.
        :return: dict
        """
        return self._default_fields

    def get_field_by_name(self, name):
        """
        Returns the true field name of a jira field based on its friendly name
        :param name: The friendly name of the field
        :return: A string with the true name of a field.
        """
        assert self.fields

        try:
            _name = name.lower()
            if _name.lower() in self.fields:
                return self.fields[_name]['id']

            else:
                return name

        except (KeyError, ValueError):
            return name

    def link_issues(self, link_type, inward, outward, comment=None):
        """
        Establishes a link in jira between two issues
        :param link_type: A string indicating the relationship from the inward to the outward
         (Example: "is part of this release")
        :param inward: Can be one of: Issue object, Issue dict, Issue key string
        :param outward: Can be one of: Issue object, Issue dict, Issue key string
        :param comment: None or a string with the comment associated with the link
        :return: No return value.
        """
        ""
        if isinstance(inward, dict):
            inward_key = inward['key']
        elif isinstance(inward, Issue):
            inward_key = inward.key
        elif isinstance(inward, str):
            inward_key = inward
        else:
            raise TypeError("'inward' parameter is not of a valid type")

        if isinstance(outward, dict):
            outward_key = outward['key']
        elif isinstance(outward, Issue):
            outward_key = outward.key
        elif isinstance(outward, str):
            outward_key = outward
        else:
            raise TypeError("'outward' parameter is not of a valid type")

        self.jira.create_issue_link(link_type, inward_key, outward_key, comment)

    def create_ticket(self, create_fields, update_fields=None, watchers=None):
        """
        Create the ticket with the required fields above.  The other keyword arguments can be used for other fields
           although the values must be in the correct format.
        :param update_fields:
        :param create_fields: All fields to include in the creation of the ticket. Keys include:
                project: A string with project key name (required)
                issuetype: A dictionary containing issuetype info (see Jira API docs) (required)
                summary: A string (required)
                description: A string
        :param update_fields: A dictionary containing reporter info  (see Jira API docs)
        :param watchers: A list of usernames that will be added to the watch list.

        :return: Return an Issue object or None if failed.
        """
        try:
            ticket = self.jira.create_issue(create_fields)

            if ticket:
                try:
                    # now update the remaining values (if any)
                    # we can't do this earlier because assignee and reporter can't be set during creation.
                    if update_fields and len(update_fields) > 0:
                        ticket.update(
                            update_fields
                        )
                except Exception as e:
                    self.logger.warning("Ticket was created but not updated due to exception: %s" % e.message)

                try:
                    if watchers and isinstance(watchers, (list, tuple)):
                        [self.jira.add_watcher(ticket, w) for w in watchers]
                except Exception as e:
                    self.logger.warning("Unable to add watcher(s) due to exception: %s" % e.message)

            return ticket

        except Exception as e:
            self.logger.error("Failed to create ticket: %s", e.message)
            return None



import logging

from dateutil.parser import parse
from munch import munchify


class InvalidId(Exception):
    pass


class InvalidData(Exception):
    pass


class JiraObject(object):
    """
    The base for all Jira objects.

    Options:
        group_id - (Optional) The ID of the group to use to understand context.  Some Jira operations will
                require some understanding of the workflow and this is provided as part of the group.
    """
    def __init__(self, source, **kwargs):
        self.source = source
        self.logger = logging.getLogger("augurjira")
        self._options = munchify(kwargs)

    def option(self, key, default=None):
        return self._options[key] if key in self._options else default

    def _set_option(self, key, value):
        self._options[key] = value
        return value

    def load(self, **kwargs):
        self._options.update(kwargs)
        return self._load()

    def _load(self):
        raise NotImplemented()

    def prepopulate(self,data):
        raise NotImplemented()

    def _convert_date_string_to_date_time(self, date_str):
        try:
            dt = parse(date_str)
            return dt
        except ValueError as e:
            self.logger.warning("JiraObject: Unable to parse string %s"%date_str)
            return None

    def log_access(self, api, *parameters):
        self.logger.error("jira:{api}:{parameters}".format(api=api, parameters=",".join([str(p) for p in parameters])))



from augur import models
from .base import ProxyBase, models_to_proxies, model_required


class StaffProxy(ProxyBase):

    def __init__(self, staff_id=None, *args, **kwargs):
        model = None
        if staff_id:
            model = models.Staff[staff_id]

        kwargs['model'] = model
        super(StaffProxy, self).__init__(*args, **kwargs)

    @property
    def _model_class(self):
        return 'Staff'


class WaypointProxy(ProxyBase):

    def __init__(self, *args, **kwargs):
        super(WaypointProxy, self).__init__(*args, **kwargs)

    @property
    def _model_class(self):
        return 'Waypoint'

    @model_required
    def get_first_state(self):
        if not self.model.states:
            return None

        initial = None
        for s in self.model.states:
            if s.is_initial_state:
                initial = s

        if not initial:
            # there is none marked as the initial state so just order by state_order then
            #   select the first one.
            initial = self.model.states.order_by(asc(WaypointStateInstance.state_order)).first()

        return initial


class JourneyProxy(ProxyBase):

    def __init__(self, journey_id=None, *args, **kwargs):
        model = None
        if journey_id:
            model = models.Journey[journed_id]

        kwargs['model'] = model,
        super(JourneyProxy, self).__init__(*args, **kwargs)

    def get_current_position(self):
        """
        Gets the current position by getting the most recent arrived at position object.
        :return: {JourneyPosition|None}
        """
        asset(self.model)
        return self.model.positions.order_by(desc(Position.arrive_time)).first()

    def post_model_create(self, model):
        # Get the first waypoint, that will be our initial position
        first_waypoint = model.path.first_waypoint

        # Once the journey has been created, create the initial
        #   position.
        position = models.JourneyPosition(
            journey=model,
            waypoint=first_waypoint,
            arrive_time=datetime.datetime.now(),
            change_type=ChangeType.first,
            waypoint_percent_complete=0)

        model.starting_position = position

        orm.commit()

    @property
    def _model_class(self):
        return 'Journey'


class PathProxy(ProxyBase):

    def __init__(self, path_id=None, *args, **kwargs):

        model = None
        if path_id:
            model = models.Path[path_id]

        kwargs['model'] = model
        super(PathProxy, self).__init__(*args, **kwargs)

    @models_to_proxies
    def start_journey(self, owner=None, artifact=Nonez):
        if not self.path:
            set_error("You must first instantiate a path before creating a journey")
            return None

        journey = JourneyProxy()
        return journey.create_model(path=self.model, owner=owner, artifacts=artifact)

    @property
    def _model_class(self):
        return 'Path'



import os
import sys

import logging

from munch import munchify

main = {}
db_instance = None


def load_settings(env=None):
    """
    (Re)Initialize global augur settings
    :return:
    """
    global main
    global db_instance

    if not env:
        env = os.environ

    main = munchify(
        {
            "project": {
                "debug": bool(int(env.get("DEBUG", True))),
                "augur_base_dir": os.path.dirname(os.path.abspath(__file__)),
                "base_dir": os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            },
            "integrations": {
                "jira": {
                    "instance": env.get("JIRA_INSTANCE",""),
                    "username": env.get("JIRA_USERNAME",""),
                    "password": env.get("JIRA_PASSWORD",""),
                    "api_path": env.get("JIRA_API_PATH", "rest/api/2"),
                },
                "confluence": {
                    "url": "%s/wiki" % env.get("CONFLUENCE_INSTANCE", env.get("JIRA_INSTANCE","")),
                    "rest_url": "%s/wiki/rest/api/content" % env.get("CONFLUENCE_INSTANCE", ""),
                    "username": env.get("CONFLUENCE_USERNAME", env.get("JIRA_USERNAME","")),
                    "password": env.get("CONFLUENCE_PASSWORD", env.get("JIRA_PASSWORD","")),
                },
                "github": {
                    "base_url": env.get("GITHUB_BASE_URL", ""),
                    "login_token": env.get("GITHUB_LOGIN_TOKEN", ""),
                    "client_id": env.get("GITHUB_CLIENT_ID", ""),
                    "client_secret": env.get("GITHUB_CLIENT_SECRET", ""),
                },
                "tempo": {
                    "api_token": env.get("TEMPO_API_TOKEN")
                }
            },
            "datastores": {
                "main": {
                    "type": env.get("DB_TYPE"),
                    "postgres": {
                        "host": env.get("POSTGRES_DB_HOST"),
                        "dbname": env.get("POSTGRES_DB_NAME"),
                        "username": env.get("POSTGRES_DB_USERNAME",""),
                        "password": env.get("POSTGRES_DB_PASSWORD", ""),
                        "port": env.get("POSTGRES_DB_PORT", ""),
                    },
                    "sqlite": {
                        "path": env.get("SQLITE_PATH"),
                    }
                },
                "cache": {
                    "mongo": {
                        "host": env.get("MONGO_HOST","localhost"),
                        "port": int(env.get("MONGO_PORT", 27017)),
                    }
                }
            }
        }
    )

    if main.project.debug:
        # set root logger to show INFO and above
        logging.getLogger().setLevel(10)

        # add handlers for loggers if they don't already have one
        if len(logging.getLogger("timer").handlers) == 0:
            logging.getLogger("timer").addHandler(logging.StreamHandler(sys.stdout))

        if len(logging.getLogger("augurjira").handlers) == 0:
            logging.getLogger("augurjira").addHandler(logging.StreamHandler(sys.stdout))

    else:
        logging.getLogger().setLevel(40)


load_settings()

